/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Diagnostics subsystem main file
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/diagnostics.hpp>
#include <hw/diagnostics/metrics/metric.hpp>

#include <hw/report/xml_printer.hpp>

#include <hw/diagnostics/report_generator.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

Diagnostics::Diagnostics(const event_list_t &reactions, const SimulationInfo &sim_info)
{
    _reaction_units.init(reactions); // Copy reactions for analysis
    _simulation_info = sim_info;
}

void Diagnostics::apply_normal_rules()
{
    bool collapse_happened = false;
    for(uint i = 0; i < _reaction_units.size(); i++)
    {
        ReactionUnit &ru1 = _reaction_units[i];

        if(!ru1.is_collapsed())
        {
            for(uint j = 0; j < _reaction_units.size(); j++)
            {
                ReactionUnit &ru2 = _reaction_units[j];

                if(!ru2.is_collapsed())
                {
                    for(uint rule_number = 0; rule_number < MAX_NORMAL_RULE_NUMBER; rule_number++)
                    {
                        if(_error_model.check(ru1, ru2, rule_number, NORMAL_COLLAPSE) == EXACT_MATCH)
                        {
                            ReactionUnitVector rs = _error_model.apply(ru1, ru2, rule_number, NORMAL_COLLAPSE);

                            if(!rs.is_empty())
                                _reaction_units.add_reaction_units(rs);

                            collapse_happened = true;
                            break;
                        }
                    }
                }

                if(collapse_happened)
                {
                    collapse_happened = false;
                    break;
                }
            }
        }
    }
}

void Diagnostics::apply_fuzzy_rules()
{
    for(uint i = 0; i < _reaction_units.size(); i++)
    {
        ReactionUnit &ru1 = _reaction_units[i];

        if(!ru1.is_collapsed())
        {
            int metric = WORST_MATCH;
            int selected_rule_number = EMPTY_RULE;
            int ru2_number;

            for(uint j = 0; j < _reaction_units.size(); j++)
            {
                ReactionUnit &ru2 = _reaction_units[j];

                if(!ru2.is_collapsed())
                {
                    for(uint rule_number = 0; rule_number < MAX_FUZZY_RULE_NUMBER; rule_number++)
                    {
                        int metric_tmp = _error_model.check(ru1, ru2, rule_number, FUZZY_COLLAPSE);

                        if(metric_tmp < metric)
                        {
                            metric = metric_tmp;
                            selected_rule_number = rule_number;
                            ru2_number = j;
                        }
                    }
                }
            }

            if(metric < WORST_MATCH)
            {
                ReactionUnit &ru2 = _reaction_units[ru2_number];
                ReactionUnitVector rs = _error_model.apply(ru1, ru2, selected_rule_number, FUZZY_COLLAPSE);
                if(!rs.is_empty())
                    _reaction_units.add_reaction_units(rs);

            }
        }
    }
}

void Diagnostics::optional_conversion()
{
    for(uint i = 0; i < _reaction_units.size(); i++)
    {
        ReactionUnit &ru1 = _reaction_units[i];
        cpptesk::hw::Reaction &r1 = ru1.getReaction();
        
        if(r1.getError() == NORMAL_REACTION)
        {
            if(r1.getSpecMessage() == NULL && r1.getImplMessage()->isOptional())
                r1.setError(UNEXPECTED_REACTION);
            else if(r1.getImplMessage() == NULL && r1.getSpecMessage()->isOptional())
                r1.setError(MISSING_REACTION);
            else if(r1.getImplMessage() != NULL && r1.getSpecMessage() != NULL && 
            (r1.getSpecMessage()->isOptional() || r1.getImplMessage()->isOptional()) &&
            Rule::compare(r1.getSpecMessage(), r1.getImplMessage()) != EXACT_MATCH)
                r1.setError(INCORRECT_REACTION);
        }
    }
}

ReportUnit Diagnostics::run()
{
    if(_reaction_units.is_empty())
        CPPTESK_DEBUG_PRINT(cpptesk::hw::DEBUG_TOOL, "Reaction set is empty!");

    // Convert optional reactions to error reactions
    optional_conversion();

    // Apply rules
    apply_normal_rules();
    apply_fuzzy_rules();

    // Print all reactions in log file
    //cpptesk::hw::report::ReactionPrinter reaction_printer = cpptesk::hw::report::ReactionPrinter();
    //reaction_printer.print((char*)"diagnostic_trace.txt", _reaction_units);
    
    ReportUnit report_unit;
    report_unit.reaction_vector = _reaction_units;
    
    // Create report
    ReportGenerator report_generator = ReportGenerator(_reaction_units, _simulation_info);
    cpptesk::hw::report::Report report = report_generator.run();
 
    report_unit.report = report;
    
    return report_unit;
}

}}}
