/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Failure finding facilities
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/error_model.hpp>

#include <hw/diagnostics/rules/rule1.hpp>
#include <hw/diagnostics/rules/rule2.hpp>
#include <hw/diagnostics/rules/rule3.hpp>
#include <hw/diagnostics/rules/rule4.hpp>
#include <hw/diagnostics/rules/rule5.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {
    
ErrorModel::ErrorModel()
{
    Rule *rule[5];

    rule[0] = new Rule1(); // Remove NORMAL reactions
    rule[1] = new Rule2(); // Recombine two INCORRECT reactions (two modes)
    rule[2] = new Rule3(); // Recombine MISSING and UNEXPECTED reactions
    rule[3] = new Rule4(); // Recombine INCORRECT and MISSING reactions
    rule[4] = new Rule5(); // Recombine INCORRECT and UNEXPECTED reactions

    for(int i = 0; i < 5; i++)
    {   
        _rules.push_back(rule[i]);
    }
}

ErrorModel::~ErrorModel()
{
    // Remove objects-rules
    for(unsigned int i = 0; i < _rules.size(); i++)
        delete(_rules[i]);
}

int ErrorModel::check(cpptesk::hw::Reaction &r1, cpptesk::hw::Reaction &r2, uint rule_number, CollapseType type)
{
    if((r1.getSpecMessage() == NULL && r1.getImplMessage() == NULL) ||
       (r2.getSpecMessage() == NULL && r2.getImplMessage() == NULL))
    {
        return FALSE_MATCH;
    }
    else
    {
        CPPTESK_ASSERTION(rule_number < (type == NORMAL_COLLAPSE ? MAX_NORMAL_RULE_NUMBER :
                                                                   MAX_FUZZY_RULE_NUMBER), "Too big rule_number");
        if(type == NORMAL_COLLAPSE)
            switch(rule_number)
            {
                case 1: return ((Rule2*)_rules[1])->check_completely(r1, r2); // to receive two normals
                case 5: return _rules[1]->check(r1, r2);                      // to receive at least one normal
                default: break;
            }
        return _rules[rule_number]->check(r1, r2);
    }
}

int ErrorModel::check(ReactionUnit &ru1, ReactionUnit &ru2, uint rule_number, CollapseType type)
{
    cpptesk::hw::Reaction &r1 = ru1.getReaction();
    cpptesk::hw::Reaction &r2 = ru2.getReaction();
    return check(r1, r2, rule_number, type);
}

ReactionUnitVector ErrorModel::apply(ReactionUnit &ru1, ReactionUnit &ru2, uint rule_number, CollapseType type)
{
    (_rules[rule_number])->assign_rule_number(ru1, ru2, rule_number, type);

    if(type == NORMAL_COLLAPSE && rule_number == 5)
        return (_rules[1])->apply(ru1, ru2);

    return (_rules[rule_number])->apply(ru1, ru2);
}

/*std::string ErrorModel::toString(uint rule_number)
{
    CPPTESK_ASSERTION(rule_number < _rules.size(), "incorrect rule number");
    return _rules[rule_number]->toString();
};*/

}}} // namespace cpptesk::hw::diagnostics
