/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Hamming Metric for diagnostics rules
///////////////////////////////////////////////////////////////////////////////

#include <hw/diagnostics/metrics/hamming_metric.hpp>

#include <math.h>

namespace cpptesk {
namespace hw {
namespace diagnostics {
        
/// Return number of identical bits
int HammingMetric::compare(const cpptesk::hw::Message &m1, const cpptesk::hw::Message &m2)
{
   int result = FALSE_MATCH;

    if(m1.size() == m2.size())
    {
        if(m1.size() == 0)
        {
            result = EXACT_MATCH;
        }
        else
        {
            uint closeness = 0; // closeness between two messages
            unsigned width = 0;
            
            for(cpptesk::hw::Message::const_iterator i = m1.begin(),
                                                     j = m2.begin();
                i != m1.end(); i++, j++)
            {
                const Field<> &field_from_m1 = m1.**i;
                const Field<> &field_from_m2 = m2.**j;
                
                if(field_from_m1.isComparable())
                {
                    if(field_from_m1.size() > 1)
                    {
                        for(unsigned j = 0; j < field_from_m1.size(); j++)
                        {
                            int a = field_from_m1.get(j);
                            int b = field_from_m2.get(j);
                            unsigned a_width = field_from_m1.get_width();
                            
                            width = width + a_width;
                            closeness = closeness + (a_width - hamming_size(a, b));
                        }
                    }
                    else
                    {
                        int a = field_from_m1.get();
                        int b = field_from_m2.get();
                        unsigned a_width = field_from_m1.get_width();
                        
                        width = width + a_width;
                        closeness = closeness + (a_width - hamming_size(a, b));
                    }
                }
            }
            
            if(width == 0) return 0;
            else result = (WORST_MATCH * (width - closeness)) / width;
        }
    }
    return result;
}

HammingMetric* HammingMetric::getMetric()
{
    return cpptesk::hw::diagnostics::SingleMetric<HammingMetric>::getInstance();
}

/// Return Hamming size
int HammingMetric::hamming_size(int a, int b)
{
    int c = a xor b;
    int counter = 0;
    
    while(c != 0)
    {
        counter++;
        
        c = (c-1)&c;
    }
    return counter;
}

}}}
