/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Generate Diagnostic Report
///////////////////////////////////////////////////////////////////////////////

#include <iomanip>
#include <stdio.h>

#include <hw/component.hpp>
#include <hw/interface.hpp>
#include <hw/message.hpp>

#include <hw/diagnostics/report_generator.hpp>

namespace cpptesk {
namespace hw {
namespace diagnostics {

ReportGenerator::ReportGenerator(ReactionUnitVector &ruv, SimulationInfo &sim_info) : 
                                _reaction_units(ruv), _sim_info(sim_info)
{
    _reaction_timeout = sim_info.reaction_timeout;
    _unexpected_reaction_timeout = sim_info.unexpected_reaction_timeout;
    _prepare_reaction_pairs();
    _expected_missing = sim_info.expected_missing;//_get_expected_missing();
    _optional_missing = sim_info.optional_missing;//_get_optional_missing();
}

ReportGenerator::~ReportGenerator()
{
}

void ReportGenerator::_prepare_reaction_pairs()
{
    for(uint i = 0; i < _reaction_units.size(); i++)
    {
        ReactionUnit &ru = _reaction_units[i];
        int parent1_id = ru.get_parent_spec_id(),
            parent2_id = ru.get_parent_impl_id();
        Reaction &r = ru.getReaction();

        int rule_number = ru.get_collapse_rule();
        
        if((!ru.is_collapsed() || r.getError() == NORMAL_REACTION) &&
           !(parent1_id == -1 && parent2_id == -1 && r.getError() == NORMAL_REACTION) &&    /* Don't process normal reactions from test oracle */
           !(r.getError() == MISSING_REACTION && r.isExpected()) &&                         /* Don't process normal missing reactions from test oracle */
           !(r.getError() == MISSING_REACTION && r.getSpecMessage()->isOptional()) &&       /* Don't process optional missing reactions */
           !(r.getError() == UNEXPECTED_REACTION && r.getImplMessage()->isOptional()) &&    /* Don't process optional unecpected reactions */
           !((r.getError() == INCORRECT_REACTION || r.getError() == NORMAL_REACTION) && r.getImplMessage()->isOptional() && 
           r.getSpecMessage()->isOptional()))                                                 /* Don't process optional incorrect reactions */
        {
            ReactionPair reaction_pair;
            reaction_pair.add_collapsed_reaction(ru.getReaction());

            reaction_pair.set_collapsed(false);
            
            if(parent1_id != -1)
            {
                while(_reaction_units[parent1_id].get_parent_spec_id() != -1)
                    parent1_id = _reaction_units[parent1_id].get_parent_spec_id();
                reaction_pair.add_spec_reaction(_reaction_units[parent1_id].getReaction());
                
                reaction_pair.set_collapsed(true);
            }

            if(parent2_id != -1)
            {
                while(_reaction_units[parent2_id].get_parent_impl_id() != -1)
                    parent2_id = _reaction_units[parent2_id].get_parent_impl_id();
                reaction_pair.add_impl_reaction(_reaction_units[parent2_id].getReaction());
                
                reaction_pair.set_collapsed(true);
            }

            reaction_pair.process(_reaction_timeout, _unexpected_reaction_timeout);

            bool incorrect_order = false;
            if(parent2_id != -1 && parent1_id != -1)
            {
                if (reaction_pair.get_spec_reaction().getError() == INCORRECT_REACTION && 
                    reaction_pair.get_impl_reaction().getError() == INCORRECT_REACTION)
                incorrect_order = true;
            }
            
            if(reaction_pair.get_group() != NONE_ERROR || incorrect_order ||
                rule_number == 1 || rule_number == 5 || 
                r.getError() == MISSING_REACTION || r.getError() == UNEXPECTED_REACTION)
                _reaction_pairs.add_reaction_pair(reaction_pair);
        }
    }
}

/*
uint ReportGenerator::_get_expected_missing()
{
    uint result = 0;
    
    for(uint i = 0; i < _reaction_units.size(); i++)
    {
        if(_reaction_units[i].getReaction().isExpected() &&
           _reaction_units[i].getReaction().getError() == MISSING_REACTION)
            result++;
    }
    return result;
}

uint ReportGenerator::_get_optional_missing()
{
    uint result = 0;
    
    for(uint i = 0; i < _reaction_units.size(); i++)
    {
        Reaction &_reaction = _reaction_units[i].getReaction();

        if(_reaction.getSpecMessage()->isOptional() && 
            !_reaction_units[i].getReaction().isExpected() &&
           _reaction.getError() == MISSING_REACTION && !_reaction_units[i].is_collapsed())
            result++;
    }
    return result;
}
*/

cpptesk::hw::report::Report ReportGenerator::run()
{
    _msg_size = 80;
    
    if(!_reaction_pairs.is_empty())
    {
        int error_num = 0;
        for(int j = 1; j >= 0; j--)
            for(int i = DATA_TIME_IFACE_ERROR; i >= NONE_ERROR; i--)
            {
                _add_error(j == 1 ? true : false, error_num, i, INCORRECT_REACTION, NORMAL_REACTION);
                _add_error(j == 1 ? true : false, error_num, i, MISSING_REACTION);
                _add_error(j == 1 ? true : false, error_num, i, UNEXPECTED_REACTION);
            }
    }

    _add_result(_sim_info.stimuli, _sim_info.reactions, _sim_info.reaction_normal,
                _sim_info.reaction_incorrect, _sim_info.reaction_missing,
                _sim_info.reaction_unexpected, _sim_info.cycles, _sim_info.time);

    return _report;
}

/// Prints the results of the test
void ReportGenerator::_add_result(unsigned stimuli, unsigned reactions, unsigned reaction_normal,
                                  unsigned reaction_incorrect, unsigned reaction_missing,
                                  unsigned reaction_unexpected, uint64_t cycles, time_t time)
{
    _add_statistic(stimuli, reactions, reaction_normal, reaction_incorrect, 
                   reaction_missing, reaction_unexpected);
    _add_simulation(cycles, time);
}

/// Create statistic report
void ReportGenerator::_add_statistic(unsigned stimuli, unsigned reactions, unsigned reaction_normal,
                                     unsigned reaction_incorrect, unsigned reaction_missing,
                                     unsigned reaction_unexpected)
{
    _report.statistics.name =  "Statistics";
    
    double temp_statistic = (stimuli != 0 ? 
        ((double)reactions / (double)stimuli) : 0);

    if(reactions - reaction_normal - _expected_missing - _optional_missing == 0) // REACTIONS - NORMAL - EXPECTED MISSING
        _report.statistics.failure = false; 
    else
        _report.statistics.failure = true; 

    _report.statistics.value = temp_statistic;
    _report.statistics.unit = "(r/s)";

    _report.statistics.stimuli.name = "STIMULI";
    _report.statistics.stimuli.value = stimuli;

    _report.statistics.reactions.name = "REACTIONS";
    _report.statistics.reactions.value = reactions;

    _report.statistics.normal.name = "NORMAL";
    _report.statistics.normal.value = reaction_normal;

    _report.statistics.incorrect.name = "INCORRECT";
    _report.statistics.incorrect.value = reaction_incorrect;
    
    if(reaction_incorrect == 0)  
        _report.statistics.incorrect.failure = false; 
    else 
        _report.statistics.incorrect.failure = true; 
    
    _report.statistics.missing.name = "MISSING";
    _report.statistics.missing.value = reaction_missing;// - _expected_missing - _optional_missing;
    
    //CPPTESK_DEBUG_STREAM() << reaction_missing << "  " << _expected_missing  << "  " << _optional_missing;
    
    if(reaction_missing /*- _expected_missing - _optional_missing*/ == 0) // MISSING - EXPECTED MISSING
        _report.statistics.missing.failure = false; 
    else
        _report.statistics.missing.failure = true; 
 
    _report.statistics.expected_missing.name = "NORMAL MISSING";
    _report.statistics.expected_missing.value = _expected_missing;

    _report.statistics.optional_missing.name = "OPTIONAL MISSING";
    _report.statistics.optional_missing.value = _optional_missing;
    
    _report.statistics.unexpected.name = "UNEXPECTED";
    _report.statistics.unexpected.value = reaction_unexpected;
    if(reaction_unexpected == 0)
        _report.statistics.unexpected.failure = false; 
    else
        _report.statistics.unexpected.failure = true; 
}

/// Create simulation report 
void ReportGenerator::_add_simulation(int cycles, time_t time)
{
    double frequency = (double)cycles / (double)time;

    std::string herz = "Hz";

    if(frequency > 1000.0) { frequency /= 1000.0; herz = "KHz"; }
    if(frequency > 1000.0) { frequency /= 1000.0; herz = "MHz"; }
    if(frequency > 1000.0) { frequency /= 1000.0; herz = "GHz"; }

    _report.simulation.name = "Simulation";

    _report.simulation.cycle.name = "Cycles";
    _report.simulation.cycle.value = cycles;
    _report.simulation.cycle.unit = "cycle(s)";

    _report.simulation.time.name = "Time";
    _report.simulation.time.value = time;
    _report.simulation.time.unit = "sec(s)";

    _report.simulation.frequency.name = "Frequency";
    _report.simulation.frequency.value = frequency;
    _report.simulation.frequency.unit = herz;
}

int ReportGenerator::_get_number_of_this_failure(bool collapsed, int count, Reaction main_reaction, int error_type_one,
                                             int error_type_two)
{
    int failure_size = 0;

    for(int j = 0; j < _reaction_pairs.size(); j++ )
    {
        Reaction &reaction = _reaction_pairs[j].get_collapsed_reaction();

        const ::cpptesk::hw::Interface *iface      = reaction.where();
        const ::cpptesk::hw::Interface *main_iface = main_reaction.where();

        std::string iface_name_00 = _get_iface_array_name(iface);
        std::string iface_name_01 = _get_iface_array_name(main_iface);

        if((_reaction_pairs[j].get_group() == count) && (_reaction_pairs[j].is_collapsed() == collapsed) &&
            ((reaction.getError() == error_type_one) || (reaction.getError() == error_type_two))&&
            (iface_name_00 == iface_name_01))
            failure_size ++;
    }

    return failure_size;
}

void ReportGenerator::_add_error(bool collapsed, int &error_num, int iterator, int error_type_one)
{
    _add_error(collapsed, error_num, iterator, error_type_one, error_type_one);
}

void ReportGenerator::_add_error(bool collapsed, int &error_num, int errors_type, int error_type_one, int error_type_two)
{
    ReactionPair example_pair;
    int errors_size = 0;
   // int failure_size = 0;

    std::list<Reaction> reactions;

    for(int j = 0; j < _reaction_pairs.size(); j++ )
    {
        bool double_iface = false;

        Reaction &reaction = _reaction_pairs[j].get_collapsed_reaction();
        
        if((_reaction_pairs[j].get_group() == errors_type) && (_reaction_pairs[j].is_collapsed() == collapsed) &&
            ((reaction.getError() == error_type_one) || (reaction.getError() == error_type_two)))
        {
            if(reactions.size() == 0)
            {
                // Add multiplicity
                _failure.value = _get_number_of_this_failure(collapsed, errors_type, reaction, error_type_one, error_type_two);

                reactions.push_back(reaction);
                _add_error_group(error_num, _reaction_pairs[j], errors_type);

                // Add explanation for spec reactions
                if(reaction.getError() != UNEXPECTED_REACTION)
                    _failure.info = reaction.getSpecMessage()->getExplanation();
             }
            else
            {
                for (std::list<Reaction>::iterator i = reactions.begin(); i != reactions.end(); i++)
                {
                    const ::cpptesk::hw::Interface *iface      = reaction.where();
                    const ::cpptesk::hw::Interface *iface_list = ((Reaction)*i).where();

                    std::string iface_name_00 = _get_iface_array_name(iface);
                    std::string iface_name_01 = _get_iface_array_name(iface_list);

                    if(iface_name_00 == iface_name_01) 
                    {
                        double_iface = true;
                        break;
                    }
                }

                if(!double_iface)
                {
                    // Add multiplicity
                    _failure.value = _get_number_of_this_failure(collapsed, errors_type, reaction, error_type_one, error_type_two);

                    reactions.push_back(reaction);
                    _add_error_group(error_num, _reaction_pairs[j], errors_type);
                    
                    // Add explanation for spec reactions
                    if(reaction.getError() != UNEXPECTED_REACTION)
                        _failure.info = reaction.getSpecMessage()->getExplanation();
                }
            }

            errors_size ++;

            if(reactions.size() == 0 || !double_iface)
            {
                _failure.data.failure  = (errors_type & DATA_ERROR)  != 0;
                _failure.time.failure  = (errors_type & TIME_ERROR)  != 0;
                _failure.iface.failure = (errors_type & IFACE_ERROR) != 0;

                 cpptesk::hw::report::FailureGroup failure_group;
                 failure_group.add(_failure);
                
                _report.failures.add(failure_group);
                
                cpptesk::hw::report::Failure failure;
                _failure = failure;
            }
        }
    }

    reactions.clear();
}

void ReportGenerator::_add_error_group(int &error_num, ReactionPair example_pair, const int group)
{
    error_num ++;

    // Add failure number
    std::stringstream temp;
    temp << error_num;
    _failure.name = temp.str();
   
    Reaction &comp_reaction = example_pair.get_collapsed_reaction();

    if(!example_pair.is_collapsed())
    {
        const ::cpptesk::hw::Interface *iface = comp_reaction.where();

        // Print the table title
        _add_title_error(group, iface, comp_reaction, comp_reaction.getError(), example_pair.is_collapsed());

        if(comp_reaction.getError() == MISSING_REACTION)
        {
            // Add fields of missing reaction
            _add_fields(true, false, comp_reaction, false); // spec, impl, reaction
        }
        else if (comp_reaction.getError() == UNEXPECTED_REACTION)
        {
            // Add fields of unexpected reaction
            _add_fields(false, true, comp_reaction, false);
        }
        else if ((comp_reaction.getError() == INCORRECT_REACTION) || (comp_reaction.getError() == NORMAL_REACTION))
        {
            bool nothing = true;

            if((group & DATA_ERROR) != 0)
            {
                _add_incorrect_data_error(comp_reaction);
                nothing = false;
            }

            if((group & TIME_ERROR) != 0)
            {
                _failure.time.spec = _get_time_interval(comp_reaction);

                nothing = false;
            }

            if((group & IFACE_ERROR) != 0)
            {
                std::string str;
                std::stringstream o_str;
                o_str << ((Component*)iface)->getName();
                o_str >> str;
                
                _failure.iface.spec = str;

                nothing = false;
            }

            if(nothing)
            {
                _add_correct_data(comp_reaction);
            }
        }
    }
    else
    {
        Reaction comp_reaction = example_pair.get_collapsed_reaction();

        if(comp_reaction.getError() == MISSING_REACTION)
        {
            // Print the table title
            const ::cpptesk::hw::Interface *iface = comp_reaction.where();

            _add_title_error(group, iface, comp_reaction, comp_reaction.getError(), example_pair.is_collapsed());

            _add_fields(true, false, comp_reaction, false); // spec, impl, reaction

        }
        else if(comp_reaction.getError() == UNEXPECTED_REACTION)
        {
            const ::cpptesk::hw::Interface *iface = comp_reaction.where();

            _add_title_error(group, iface, comp_reaction, comp_reaction.getError(), example_pair.is_collapsed());

            _add_fields(false, true, comp_reaction, false); // spec, impl, reaction

        }
        else // NORMAL or INCORRECT
        {
            Reaction spec_reaction = example_pair.get_spec_reaction();
            Reaction impl_reaction = example_pair.get_impl_reaction();

            _add_title_error(group, impl_reaction.where(), impl_reaction, comp_reaction.getError(), example_pair.is_collapsed());

            if((group & DATA_ERROR) != 0)
                _add_incorrect_data_error(comp_reaction);
            else
                _add_correct_data(comp_reaction);

            if((group & TIME_ERROR) != 0)
            {
                _failure.time.spec = _get_time_interval(spec_reaction);
            }

            if((group & IFACE_ERROR) != 0)
            {
                std::string str;
                std::stringstream o_str;
                o_str << ((Component*)spec_reaction.where())->getName();
                o_str >> str;
                
                _failure.iface.spec = str;
            }
        }
    }
}

void ReportGenerator::_add_fields(bool spec_get, bool impl_get, Reaction comp_reaction, bool error)
{
    const ::cpptesk::hw::Message *message_spec;
    const ::cpptesk::hw::Message *message_impl;

    if(spec_get) message_spec = comp_reaction.getSpecMessage();
    if(impl_get) message_impl = comp_reaction.getImplMessage();
    
    cpptesk::hw::Message::const_iterator j_message;
    
    if(spec_get & impl_get)
        j_message = message_impl->begin();
    
    _failure.data.name = (spec_get ? message_spec:message_impl)->getName();
    
    for(cpptesk::hw::Message::const_iterator i_message = (spec_get ? message_spec:message_impl)->begin();
        i_message != (spec_get ? message_spec:message_impl)->end(); 
        i_message++)
    {
        cpptesk::hw::report::FieldFailure report_field;

        bool write_flag = true;
        
        if(spec_get)
        {
            std::stringstream buffer;
            std::string str_spec;
            
            const Field<> &field = *message_spec.**i_message;
            
            std::string str_name = std::string(field.getShortName());
            
            if(field.size() > 1)
            {
                if(impl_get && error)
                {
                    const Field<> &field_impl = *message_impl.**j_message;
                    
                    std::string str_impl;
                    
                    for(unsigned i = 0; i < field_impl.size(); i++)
                        if(field_impl.get(i) != field.get(i))
                        {
                            buffer << std::hex << field_impl.get(i);
                            buffer >> str_impl;

                            buffer.clear();
                            buffer << std::hex << field.get(i);
                            buffer >> str_spec;
                            
                            buffer.clear();
                            std::string str_temp;
                            buffer << "["<< std::dec << i << "]";
                            buffer >> str_temp;
                            str_name.append(str_temp);
                            break;
                        }

                    _add_fields(report_field.impl, str_impl);
                        
                    write_flag = false;
                }
                else
                    str_name.append("[0]");
            }
            report_field.name = str_name;
            
            if(write_flag)
            {
                if(field.size() > 1)
                    buffer << std::hex << field.get(0);
                else
                    buffer << std::hex << field.get();
                
                buffer >> str_spec;
            }
            
            _add_fields(report_field.spec, str_spec);
            
            report_field.comparable = field.isComparable();
        }
        
        if(impl_get && write_flag)
        {
            std::string str_name;
            
            std::stringstream buffer;

            const Field<> &field = *message_impl.**(spec_get ? j_message:i_message);

            if(!spec_get)
            {
                str_name = std::string(field.getShortName());
                if(field.size() > 1)
                    str_name.append("[0]");
                report_field.name = str_name;

                report_field.comparable = field.isComparable();
                str_name.clear();
            }
            
            if(field.size() > 1)
                buffer << std::hex << field.get(0);
            else
                buffer << std::hex << field.get();
            
            buffer >> str_name;
            
            _add_fields(report_field.impl, str_name);
        }

        if(spec_get & impl_get)
        {
            j_message++;
            
            cpptesk::hw::report::FieldList::iterator i = report_field.spec.end();
            cpptesk::hw::report::FieldList::iterator j = report_field.impl.end();
            
            do
            {
                i--;
                j--;

                cpptesk::hw::report::ItemField &item_spec = *i;
                cpptesk::hw::report::ItemField &item_impl = *j;
                
                if(item_impl.value != item_spec.value)
                {
                    item_impl.failure = true;
                    item_spec.failure = true;
                    
                    if(!report_field.failure) report_field.failure = true;
                }
            }
            while(j != report_field.impl.begin() && i != report_field.spec.begin());
            
            while(j != report_field.impl.begin())
            {
                j--;

                cpptesk::hw::report::ItemField &item_impl = *j;

                item_impl.failure = true;
                
                if(!report_field.failure) report_field.failure = true;
            }
            
            while(i != report_field.spec.begin())
            {
                i--;

                cpptesk::hw::report::ItemField &item_spec = *i;

                item_spec.failure = true;
                
                if(!report_field.failure) report_field.failure = true;
            }
            
        }
            
        _failure.data.add(report_field);
    }
}

// Add field items to FieldList
void ReportGenerator::_add_fields(cpptesk::hw::report::FieldList &field_list, std::string field_string)
{
    for(int i = 0; i < (int)field_string.length(); i++)
    {
        cpptesk::hw::report::ItemField temp_item;
        temp_item.value = field_string.substr(i, 1);
        temp_item.failure = false;
        field_list.add(temp_item);
    }
}

// Print DATA for NORMAL reactions 
void ReportGenerator::_add_correct_data(Reaction comp_reaction)
{
    _add_fields(true, true, comp_reaction, false); // spec, impl, reaction
}

void ReportGenerator::_add_incorrect_data_error(Reaction comp_reaction)
{
    _add_fields(true, true, comp_reaction, true); // spec, impl, reaction
}

std::string ReportGenerator::_get_time_interval(Reaction reaction)
{
    std::stringstream time_interval;

    // Get interface of reaction
    const ::cpptesk::hw::Interface *iface = reaction.where(); 
    
    return _get_time_interval(reaction, iface);
}

std::string ReportGenerator::_get_time_interval(Reaction reaction, const ::cpptesk::hw::Interface *iface)
{
    std::stringstream time_interval;

    // Get dTi
    int impl_time_reaction;
    if(iface->isUnexpectedReactionTimeoutValid())
    {
        impl_time_reaction = iface->getUnexpectedReactionTimeout();    
    }
    else
        impl_time_reaction = _unexpected_reaction_timeout;
    
    if(impl_time_reaction > 0)
    {
        impl_time_reaction = reaction.whenStarted() - impl_time_reaction;
        
        if(impl_time_reaction < 0) impl_time_reaction = 0;
        
        time_interval << std::dec << impl_time_reaction << "..";
    }
    
    time_interval << std::dec << "(" << reaction.whenStarted() << ")";
    
    // Get dTs
    int spec_time_reaction;
    if(iface->isReactionTimeoutValid())
    {
        spec_time_reaction = iface->getReactionTimeout();
    }
    else
        spec_time_reaction = _reaction_timeout;
    
    if(spec_time_reaction > 0)
    {
        spec_time_reaction = reaction.whenStarted() + spec_time_reaction;
        time_interval << std::dec << ".." << spec_time_reaction;
    }
    
    std::string time_str;
    time_interval >> time_str;
    return time_str;
}

std::string ReportGenerator::_get_iface_array_name(const ::cpptesk::hw::Interface *iface)
{
    std::string str;
    std::stringstream o_str;
    o_str << ((Component*)iface)->getName();
    o_str >> str;

    unsigned found_0 = str.find_first_of("[");
    unsigned found_1 = str.find_last_of("[");

    if(found_0 == found_1)
    {
        return str;
    }
    else
    {
        return str.substr(0,found_1);
    }
}

void ReportGenerator::_add_title_error(int errors_type, const ::cpptesk::hw::Interface *iface_impl, 
                                       Reaction &reaction, error_type_t global_error_type, bool collapsed)
{
    std::stringstream o_str;
    std::string str_iface_name;
    std::string str_cycle;

    o_str << ((Component*)iface_impl)->getName();
    o_str >> str_iface_name;
    o_str.clear();

    _failure.iface.name = str_iface_name; // Add iface name
    _failure.iface.failure = (errors_type & IFACE_ERROR) != 0;

    if(global_error_type == MISSING_REACTION)
        o_str << _get_time_interval(reaction, iface_impl);
    else
        o_str << std::dec << reaction.whenStarted();
    
    o_str >> str_cycle;

    _failure.time.value = str_cycle;
    _failure.time.failure = (errors_type & TIME_ERROR) != 0;

    if(global_error_type != UNEXPECTED_REACTION)
    {
        if(global_error_type != MISSING_REACTION)
        {
            // INCORRECT DATA
            if((errors_type & DATA_ERROR) != 0)
            {
                _failure.type = "INCORRECT";
            }
            else
                if((errors_type & IFACE_ERROR) != 0)
                {
                    _failure.type = "IFACE VIOLATION";
                }
                else 
                    if((errors_type & TIME_ERROR) != 0)
                    {
                        _failure.type = "TIMING VIOLATION";
                    }
                    else
                    {
                        _failure.type = "INCORRECT ORDER";
                    }
        }
        else
        {
            _failure.type = "MISSING";
        }
    }
    else
    {
        _failure.type = "UNEXPECTED";
    }

    if(collapsed)
    {
        _failure.type.append("?");
    }
}
}}} 
