/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: History of events (stimuli & reactions)
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/event.hpp>
#include <hw/errors.hpp>
#include <hw/diagnostics/metrics/metric.hpp>

#ifdef CPPTESK_BUILD_WITH_OLD_ASPECTRACE_PATH
extern "C" {
  #include <aspectrace.h>
  #include <aspect_coverage.h>
  #include <aspect_error.h>
}
#else
extern "C" {
  #include <aspectrace/aspectrace.h>
  #include <aspectrace/aspect_coverage.h>
  #include <aspectrace/aspect_error.h>
}
#endif // CPPTESK_BUILD_WITH_OLD_ASPECTRACE_PATH

#include <list>
#include <string>

namespace cpptesk {
namespace hw {

enum error_mode_t
{
    LAST_ERROR    = 0x1,
    LAST_EXPECTED = 0x2,
    ALL_ERRORS    = 0x4,
    ALL_EXPECTED  = 0x8
};

class EventHistory
{
public:
    const static size_t HISTORY_SIZE = 1024;

    EventHistory();
    virtual ~EventHistory();

    /// Registers the stimulus and returns the shared pointer to the event.
    event_ptr_t register_stimulus(const Stimulus &stimulus);
    /// Registers the reaction and returns the shared pointer to the event.
    event_ptr_t register_reaction(const Reaction &reaction);

    /// Removes all reactions from the current set and appends them to the history.
    void reset_current_set();

    /// Returns false iff there is at least one errornous reaction in the current set.
    /// Removes reactions from the current set and appends them to the history.
    bool verdict();

    /// Returns an error message or a concatenation of several messages.
    std::string error(error_mode_t = LAST_ERROR) const;

    /// Returns a number of reactions (errors) of a given type.
    unsigned count(error_type_t error) const;

    /// Removes all stimuli and reactions including the history and resets statistics.
    void reset();

    /// Traces statistics on stimuli and reactions.
    void trace_statistics();

    /// Traces human-understandable diagnostics for reactions.
    void trace_diagnosis(uint64_t cycles, time_t time);

    void setReactionTimeout(cycle_t timeout);
    void setUnexpectedReactionTimeout(cycle_t timeout);

protected:
    typedef event_list_t stimulus_list_t;
    typedef event_list_t reaction_list_t;

    unsigned _id;

    /// Current set of reactions.
    reaction_list_t _reactions;
    /// Current set of stimuli.
    stimulus_list_t _stimuli;

    /// History of reactions.
    reaction_list_t _history;
    
    /// Stimuli statistics.
    unsigned _stimulus_count;

    /// Reactions statistics.
    unsigned _reaction_count[__ERROR_TYPE_NUMBER];
    unsigned _optional_missing;
    unsigned _expected_missing;

    cycle_t _reaction_timeout;
    cycle_t _unexpected_reaction_timeout;
};

}} // namespace cpptesk::hw
