/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Macros for calling methods/processes
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/debug.hpp>

////////////////////////////////////////////////////////////////////////////////
// Auxiliary macros
////////////////////////////////////////////////////////////////////////////////

#define __CPPTESK_PROCESS_PTR() ((::cpptesk::hw::Process*)__cpptesk_process)

#define __CPPTESK_CHECK_IFACE(iface) \
    ::cpptesk::hw::__cpptesk_assertion_file_line((iface).isReady(), \
        "Interface " + (iface).getName() + " has been already requested", __FILE__, __LINE__)

#define __CPPTESK_CHECK_START_MODE(model, full_method_name, mode) \
    ::cpptesk::hw::__cpptesk_assertion_file_line((model).inside_cycle() || \
        (mode & ::cpptesk::hw::EMPTY_MODEL::START_THIS_CYCLE) == 0, \
        "Cannot start " #full_method_name " between two cycles", __FILE__, __LINE__)

#define __CPPTESK_PREPARE_CALL(method_name) \
    prepare_call(::cpptesk::hw::Process::CallDescriptor(#method_name, __FILE__, __LINE__))

////////////////////////////////////////////////////////////////////////////////
// Calling local methods/processes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_CALL_PARALLEL
 * Creates a process for the method and starts it immediately.
 */
#define CPPTESK_CALL_PARALLEL(method_name, iface, message) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line(isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        /* IMPORTANT: Do not check existance of START() */ \
        __CPPTESK_PREPARE_CALL(method_name), \
        start((model_t::method_t)&model_t::method_name, iface, message, __CPPTESK_PROCESS_PTR(), START_IMMEDIATELY) \
    )

/** @def CPPTESK_CALL_SEQUENTIAL
 * Invokes the method within the current process.
 */
#define CPPTESK_CALL_SEQUENTIAL(method_name, iface, message) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line(isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        __CPPTESK_CHECK_START(), \
        __CPPTESK_PREPARE_CALL(method_name), \
        start((model_t::method_t)&model_t::method_name, CPPTESK_GET_PROCESS(), iface, message) \
    )

/** @def CPPTESK_CALL_PARALLEL_EX
 * Creates a process and starts it according the mode (full method name is used).
 */
#define CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line(isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        /* IMPORTANT: Do not check existence of START() */ \
        __CPPTESK_PREPARE_CALL(full_method_name), \
        start((model_t::method_t)&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), mode) \
    )

/** @def CPPTESK_CALL_SEQUENTIAL_EX
 * Invokes the method within the current process (full method name is used).
 */
#define CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line(isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        /* IMPORANT: Do not check existence of START() */ \
        __CPPTESK_PREPARE_CALL(full_method_name), \
        start((model_t::method_t)&full_method_name, CPPTESK_GET_PROCESS(), iface, message) \
    )

/** @def CPPTESK_CALL_PROCESS
 * Creates a process for the method or executes the method within the current process:
 *   PARALLEL   => CPPTESK_CALL_PARALLEL(...)
 *   SEQUENTIAL => CPPTESK_CALL_SEQUENTIAL(...)
 */
#define CPPTESK_CALL_PROCESS(call_mode, method_name, iface, message) \
    ((call_mode) == PARALLEL ? \
        CPPTESK_CALL_PARALLEL(method_name, iface, message) : \
        CPPTESK_CALL_SEQUENTIAL(method_name, iface, message))

/** @def CPPTESK_CALL_PROCESS_EX
 * Creates a process for the method or executes the method within the current process:
 *   PARALLEL   => CPPTESK_CALL_PARALLEL_EX(...)
 *   SEQUENTIAL => CPPTESK_CALL_SEQUENTIAL_EX(...)
 */
#define CPPTESK_CALL_PROCESS_EX(call_mode, full_method_name, iface, message, mode) \
    ((call_mode) == PARALLEL ? \
        /* IMPORTANT: Parallel call does not use the process parameter */ \
        CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, message, mode) : \
        /* IMPORTANT: Sequential call does not use the mode parameter */ \
        CPPTESK_CALL_SEQUENTIAL_EX(full_method_name, iface, message))

/** @def CPPTESK_RECV_STIMULUS
 * Executes the stimulus reception handler (in a new process or within the current process):
 *   PARALLEL   => CPPTESK_CALL_PARALLEL(<stimulus-handler>, ...)
 *   SEQUENTIAL => CPPTESK_CALL_SEQUENTIAL(<stimulus-handler>, ...)
 */
#define CPPTESK_RECV_STIMULUS(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onStimulus, iface, message)

/** @def CPPTESK_SEND_REACTION
 * Executes the reaction sending handler (in a new process or within the current process):
 *   PARALLEL   => CPPTESK_CALL_PARALLEL(<reaction-handler>, ...)
 *   SEQUENTIAL => CPPTESK_CALL_SEQUENTIAL(<reaction-handler>, ...)
 */
#define CPPTESK_SEND_REACTION(call_mode, iface, message) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
   _cpptesk_set_explanation(message, __FILE__, __LINE__))

/** @def CPPTESK_SEND_REACTION_WITH_COMMENT
 * Sets the human-readable comment for the reaction message and executes
 * the reaction sending handler (in a new process or within the current process):
 *   PARALLEL   => CPPTESK_CALL_PARALLEL(<reaction-handler>, ...)
 *   SEQUENTIAL => CPPTESK_CALL_SEQUENTIAL(<reaction-handler>, ...)
 */
#define CPPTESK_SEND_REACTION_WITH_COMMENT(call_mode, iface, message, comment) \
    CPPTESK_CALL_PROCESS(call_mode, onReaction, iface, \
    _cpptesk_set_explanation(message, __FILE__, __LINE__, comment))

/** @def CPPTESK_CALL_LOOP
 * Creates a process and launches the infinite loop (default interface and message are used).
 */
#define CPPTESK_CALL_LOOP(method_name) \
    CPPTESK_CALL_PARALLEL(method_name, CPPTESK_DEFAULT_IFACE, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE)

/** @def CPPTESK_CALL_LOOP_EX
 * Creates a process and launches the infinite loop with the given model (default message is used).
 */
#define CPPTESK_CALL_LOOP_EX(full_method_name, iface, mode) \
    CPPTESK_CALL_PARALLEL_EX(full_method_name, iface, ::cpptesk::hw::CPPTESK_DEFAULT_MESSAGE, mode)

////////////////////////////////////////////////////////////////////////////////
// Calling methods/processes of some model
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_CALL_PROCESS_OF
 * Creates a process for the method in the given model.
 */
#define CPPTESK_CALL_PROCESS_OF(model, full_method_name, iface, message) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line((model).isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* The interface is not locked */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR()) \
    )

/** @def CPPTESK_CALL_PROCESS_OF_EX
 * Creates a process for the method in the given model with the given mode.
 */
#define CPPTESK_CALL_PROCESS_OF_EX(model, full_method_name, iface, message, mode) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line((model).isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* Check the start mode */ \
        __CPPTESK_CHECK_START_MODE(model, full_method_name, mode), \
        /* Go */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), mode) \
    )

/** @def CPPTESK_CALL_STIMULUS_OF
 * Creates a process for the stimulus method in the given model.
 */
#define CPPTESK_CALL_STIMULUS_OF(model, full_method_name, iface, message) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line((model).isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        /* Trace the call */ \
        ::cpptesk::hw::Tracer::get().trace(model.time(), #model, #full_method_name, #iface, #message, message), \
        /* Prepare the call */ \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* Check the interface */ \
        __CPPTESK_CHECK_IFACE(iface), \
        /* Go and lock the interface */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), ::cpptesk::hw::EMPTY_MODEL::START_IMMEDIATELY) \
    )

/** @def CPPTESK_CALL_STIMULUS_OF_EX
 * Creates a process for the stimulus method in the given model with the given mode.
 */
#define CPPTESK_CALL_STIMULUS_OF_EX(model, full_method_name, iface, message, mode) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line((model).isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        /* Trace the call */ \
        ::cpptesk::hw::Tracer::get().trace(model.time(), #model, #full_method_name, #iface, #message, message), \
        /* Prepare the call */ \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
         /* Check the interface */ \
        __CPPTESK_CHECK_IFACE(iface), \
        /* Check the start mode */ \
        __CPPTESK_CHECK_START_MODE(model, full_method_name, mode), \
        /* Go and lock the interface */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), ::cpptesk::hw::EMPTY_MODEL::START_REQUEST_IFACE | (mode)) \
    )

/** @def CPPTESK_CALL_REACTION_OF
 * Creates a process for the reaction method in the given model.
 */
#define CPPTESK_CALL_REACTION_OF(model, full_method_name, iface, message) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line((model).isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* Go */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR()) \
    )

/** @def CPPTESK_CALL_REACTION_OF_EX
 * Creates a process for the reaction method in the given model with the given mode.
 */
#define CPPTESK_CALL_REACTION_OF_EX(model, full_method_name, iface, message, mode) \
    ( \
        cpptesk::hw::__cpptesk_assertion_file_line((model).isComponentRegistered(iface), \
          "Interface " + (iface).getName() + " has not been registered by ADD_IFACE", __FILE__, __LINE__), \
        (model).__CPPTESK_PREPARE_CALL(full_method_name), \
        /* Check the start mode */ \
        __CPPTESK_CHECK_START_MODE(model, full_method_name, mode), \
        /* Go */ \
        (model).start(&full_method_name, iface, message, __CPPTESK_PROCESS_PTR(), mode) \
    )

////////////////////////////////////////////////////////////////////////////////
// Call stack
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_CALL_STACK_OF(model) \
    ((model).process_stack())

#define CPPTESK_CALL_STACK() \
    CPPTESK_CALL_STACK_OF(*this)

////////////////////////////////////////////////////////////////////////////////
// Short names
////////////////////////////////////////////////////////////////////////////////

#ifdef HW_SHORT_NAMES
    #define CALL_PARALLEL              CPPTESK_CALL_PARALLEL
    #define CALL_SEQUENTIAL            CPPTESK_CALL_SEQUENTIAL
    #define CALL_PROCESS               CPPTESK_CALL_PROCESS
    #define CALL_PARALLEL_EX           CPPTESK_CALL_PARALLEL_EX
    #define CALL_SEQUENTIAL_EX         CPPTESK_CALL_SEQUENTIAL_EX
    #define CALL_PROCESS_EX            CPPTESK_CALL_PROCESS_EX
    #define RECV_STIMULUS              CPPTESK_RECV_STIMULUS
    #define SEND_REACTION              CPPTESK_SEND_REACTION
    #define SEND_REACTION_WITH_COMMENT CPPTESK_SEND_REACTION_WITH_COMMENT
    #define CALL_LOOP                  CPPTESK_CALL_LOOP
    #define CALL_LOOP_EX               CPPTESK_CALL_LOOP_EX
    #define CALL_PROCESS_OF            CPPTESK_CALL_PROCESS_OF
    #define CALL_PROCESS_OF_EX         CPPTESK_CALL_PROCESS_OF_EX
    #define CALL_STIMULUS_OF           CPPTESK_CALL_STIMULUS_OF
    #define CALL_STIMULUS_OF_EX        CPPTESK_CALL_STIMULUS_OF_EX
    #define CALL_REACTION_OF           CPPTESK_CALL_REACTION_OF
    #define CALL_REACTION_OF_EX        CPPTESK_CALL_REACTION_OF_EX
    #define CALL_STACK_OF(model)       CPPTESK_CALL_STACK_OF(model)
    #define CALL_STACK()               CPPTESK_CALL_STACK()
#endif // HW_SHORT_NAMES

