/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Process context switcher
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/process/process.hpp>

#if defined(WIN32)
    #include <hw/utils/win32/ucontext.h>
#elif defined(__GNUC__)
    #include <ucontext.h>
#else
    #error "unknown platform"
#endif

#include <stdint.h>
#include <stdlib.h>

namespace cpptesk {
namespace hw {

class Process;
class Interface;
class Message;
class BaseModel;

/**
 * A context switching interface.
 */
class ContextSwitcher
{
public:
    typedef void (BaseModel::*method_t)(Process &process, Interface &iface, Message &message);

    ContextSwitcher() {}
    virtual ~ContextSwitcher() {}

    /// Allocates a process context.
    virtual void get_context(Process &process) = 0;
    /// Allocates a process context and initializes it so as to executes a given method.
    virtual void make_context(BaseModel &model, method_t method, Process &process, Interface &iface, Message &message) = 0;
    /// Releases a process context.
    virtual void free_context(Process &process) = 0;
    /// Switches from one context to an other.
    virtual void swap_context(Process &from, Process &to) = 0;

protected:
    template <typename T> T* get_process_context(Process &process) { return process.get_context<T>(); }
    void set_process_context(Process &process, void *context) { process.set_context(context); }
};

/**
 * The default implementation of the context switching interface.
 * It is based on UNIX ucontext's functions: (get|make|swap)context().
 */
class DefaultContextSwitcher: public ContextSwitcher
{
public:
    DefaultContextSwitcher(): _created(0), _deleted(0) {}
    virtual ~DefaultContextSwitcher() {}

    virtual void get_context(Process &process);
    virtual void make_context(BaseModel &model, method_t method, Process &process, Interface &iface, Message &message);
    virtual void free_context(Process &process);
    virtual void swap_context(Process &from, Process &to);

private:
    const static int _PROCESS_STACK_SIZE = 32*1024;

    struct _MethodArguments
    {
        _MethodArguments(DefaultContextSwitcher &switcher, BaseModel &model, method_t method,
            Process &process, Interface &iface, Message &message):
            switcher(switcher), model(model), method(method), process(process), iface(iface), message(message) {}

        DefaultContextSwitcher &switcher;
        BaseModel              &model;
        method_t               method;
        Process                &process;
        Interface              &iface;
        Message                &message;
    };

    static void _process_wrapper(int addr_hi, int addr_lo);

    // Debug count
    unsigned _created;
    unsigned _deleted;
};

}} // namespace cpptesk::hw
