/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Printer for Diagnostics Results
///////////////////////////////////////////////////////////////////////////////

#include <iomanip>
#include <stdio.h>

#include <hw/report/console_printer.hpp>

namespace cpptesk {
namespace hw {
namespace report {

ConsolePrinter::ConsolePrinter()
{}

ConsolePrinter::~ConsolePrinter()
{}

void ConsolePrinter::print(std::ostream &out, const Report &report)
{
    print(out, report, true);
}
    
void ConsolePrinter::print(std::ostream &out, const Report &report, bool debug_color)
{
    _debug_color = debug_color;
    _line_width = 80;
    _frequency_precision = 2;

    out << std::endl; // New line
    
    _print_failures(out, report.failures);
    _print_statistic(out, report.statistics);
    _print_simulation(out, report.simulation);

    out << std::endl; // New line
}

/// Print failures from report
void ConsolePrinter::_print_failures(std::ostream &out, const cpptesk::hw::report::Failures &failures)
{
    _print_line(out); // Print separation line

    for(cpptesk::hw::report::Failures::const_iterator i = failures.begin(); i != failures.end(); i++)
    {
        const cpptesk::hw::report::FailureGroup &failure_group = *i;
    
        const cpptesk::hw::report::Failure &failure = *failure_group.begin();
        
        int failure_width = ((std::string)"Failure #").size() + ((std::string)failure.name).size();
        int data_width = ((std::string)failure.data.name).size();
        int first_column_width = failure_width > data_width ? failure_width : data_width;
        
        // Print title of failure
        _title_failures(out, failure, first_column_width);
        
        // Print failure
        
        // Print DATA title message
        std::string temp_str;
        _add_symbol(temp_str, (char*)" ", first_column_width - data_width);
        
        out << "| " << failure.data.name << temp_str << " |";

        int size_limit = _line_width - first_column_width - ((std::string)"|  |  |").size();

        for(cpptesk::hw::report::DataFailure::const_iterator j = failure.data.begin(); j != failure.data.end(); j++)
        {
            const cpptesk::hw::report::FieldFailure &field = *j;

            if((failure.data.failure && field.failure) || (!failure.data.failure))
            {
                int width = get_max_width(field.name, field.spec.toString(), field.impl.toString());
                
                temp_str.clear();

                if(size_limit - width >= 0)
                {
                    _add_symbol(temp_str, (char*)" ", width - field.name.size() + 1);
        
                    if(!field.comparable && _debug_color) out << CPPTESK_DEBUG_STYLE(YELLOW, NORMAL); 
                    
                    out << temp_str << field.name;
                    
                    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
   
                    if(size_limit - width <= 2)
                    {
                        temp_str.clear();
                        _add_symbol(temp_str, (char*)" ", size_limit - width);
                        out << temp_str << " |";

                        size_limit -= size_limit - width;
                    }
                    else
                        out << " |";     
                            
                    size_limit -= width + 3;
                }
            }
        }
        
        if(size_limit >= 0)
        {
            temp_str.clear();
            _add_symbol(temp_str, (char*)" ", size_limit + 1);
            out << temp_str;
            out << " |";     
        }

        out << std::endl;
        
        const cpptesk::hw::report::FieldFailure &first_field = *failure.data.begin();
        
        std::string temp = first_field.impl.toString();
        // Print spec data message
        if(!temp.empty()) _data_impl(out, failure, first_column_width);
        
        temp = first_field.spec.toString();;
        // Print impl data message
        if(!temp.empty()) _data_spec(out, failure, first_column_width);
        
        // Print separation line
        if(failure.time.failure || failure.iface.failure) _print_line(out);

        // Print TIME
        if(failure.time.failure) _time_failure(out, failure, first_column_width);

        // Print IFACE
        if(failure.iface.failure) _iface_failure(out, failure, first_column_width);

        // Print explanation for spec reaction
        if(failure.info.length() > 0) _print_explanation(out, failure.info);

        // Print separation line
        _print_line(out);
    }
}

/// Print spec DATA message
void ConsolePrinter::_data_spec(std::ostream &out, const cpptesk::hw::report::Failure &failure, const int &first_column_width)
{ 
    _data_line(out, failure, first_column_width, false);
}

/// Print impl DATA message
void ConsolePrinter::_data_impl(std::ostream &out, const cpptesk::hw::report::Failure &failure, const int &first_column_width)
{
    _data_line(out, failure, first_column_width, true);
}

/// Print DATA spec/impl line
inline void ConsolePrinter::_data_line(std::ostream &out, const cpptesk::hw::report::Failure &failure, const int &first_column_width, const bool &impl)
{
    std::string temp_str;
    std::string title_str = impl ? "Received" : "Expected";
    
    _add_symbol(temp_str, (char*)" ", first_column_width - title_str.size());
    out << "| " << title_str << temp_str << " | ";
    
    int size_limit = _line_width - first_column_width - ((std::string)"|  |  |").size();
        
    for(cpptesk::hw::report::DataFailure::const_iterator j = failure.data.begin(); j != failure.data.end(); j++)
    {
        const cpptesk::hw::report::FieldFailure &field = *j;

        //out << "(( " << failure.data.failure << " - failure.data.failure))" << std::endl;
        
        if((failure.data.failure && field.failure) || (!failure.data.failure))
        {
            std::string spec_str = field.spec.toString();
            std::string impl_str = field.impl.toString();
            
            int width = get_max_width(field.name, spec_str, impl_str);
            temp_str.clear();

            if((size_limit - width >= 0) && (width > 0))
            {
                _add_symbol(temp_str, (char*)" ", width - (impl ? impl_str.size() : spec_str.size()));
                out << temp_str;
                
                // impl message = RED, spec = GREEN, if not comparable
                _print_color_field(out, impl ? field.impl : field.spec, field.comparable ? (impl ? RED : GREEN) : NORMAL);
                
                if(size_limit - width <= 2)
                {
                    temp_str.clear();
                    _add_symbol(temp_str, (char*)" ", size_limit - width);
                    out << temp_str << " |";

                    size_limit -= size_limit - width;
                }
                else
                    out << " | ";     
                
                size_limit -= width + 3;
            }
        }
    }
    
    if(size_limit >= 0)
    {
        temp_str.clear();
        _add_symbol(temp_str, (char*)" ", size_limit);
        out << temp_str;
        out << " |";     
    }

    out << std::endl;     
}

/// Print time failure
void ConsolePrinter::_time_failure(std::ostream &out, const cpptesk::hw::report::Failure &failure, const int &first_column_width)
{
    std::string temp_str;
    _add_symbol(temp_str, (char*)" ", first_column_width - ((std::string)"Timing").size());

    out << "| " << "Timing" << temp_str << " | expected at ";   
    
    if(failure.time.spec.find_first_of("..") != std::string::npos) out << "[";

    if(_debug_color) out << CPPTESK_DEBUG_STYLE(GREEN, NORMAL); // To disabling color debug
    out << failure.time.spec;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL); // To disabling color debug


    if(failure.time.spec.find_first_of("..") != std::string::npos) out << "]";
    
    temp_str.clear();
    _add_symbol(temp_str, (char*)" ", _line_width - first_column_width - 2
                - ((std::string)" | expected at ").size() - failure.time.spec.size() - 3);
    out << temp_str << "|" << std::endl;
}

/// Print iface failure
void ConsolePrinter::_iface_failure(std::ostream &out, const cpptesk::hw::report::Failure &failure, const int &first_column_width)
{
    std::string temp_str;
    _add_symbol(temp_str, (char*)" ", first_column_width - ((std::string)"Interface").size());

    out << "| " << "Interface" << temp_str << " | expected on ";   
    
    out << "[";
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(GREEN, NORMAL); 
    out << failure.iface.spec;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    out << "]";
    
    temp_str.clear();
    _add_symbol(temp_str, (char*)" ", _line_width - first_column_width - 2
                - ((std::string)" | expected on ").size() - failure.iface.spec.size() - 3);
    out << temp_str << "|" << std::endl;
}

/// Print title of failure
void ConsolePrinter::_title_failures(std::ostream &out, const cpptesk::hw::report::Failure &failure, const int &first_column_width)
{
    // Print failure
    out << "| ";
    if(_debug_color) out << CPPTESK_DEBUG_STYLE( RED, NORMAL);
    
    std::string temp_str;
    _add_symbol(temp_str, (char*)" ", first_column_width - ((std::string)"Failure #").size() - ((std::string)failure.name).size());

    // Print failure number
    out << "Failure #" << failure.name << temp_str;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    
    // Print failure type: MISSING, UNEXPECTED ... etc.
    out << " | ";
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(RED, NORMAL);
    
    out << failure.type;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    out << " ";
    
    // Print failure iface
    out << "[";
    if(failure.iface.failure && _debug_color) out << CPPTESK_DEBUG_STYLE(RED, NORMAL); 
    out << failure.iface.name;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    out << "]:";
    
    // Print failure time
    if(failure.time.value.find_first_of("..") != std::string::npos)
        out << "["; 
    
    if(failure.time.failure && _debug_color) out << CPPTESK_DEBUG_STYLE(RED, NORMAL); 
    
    out << failure.time.value;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL); 
    if(failure.time.value.find_first_of("..") != std::string::npos)
        out << "]"; 
    
    // The width of output string
    int message_lenght = 2 + first_column_width + 3;
    message_lenght += ((std::string)failure.type).size() + 1 + ((std::string)failure.iface.name).size() + 3;
    if(failure.time.value.find_first_of("..") != std::string::npos)
        message_lenght +=2; 
    message_lenght += ((std::string)failure.time.value).size();
    int times_lenght = 3 + _get_size(failure.value) + ((std::string)" time(s)").size() + 2;
    
    if(_line_width - message_lenght < times_lenght)
    {
        // Add spaces befor print end of line ("|")
        std::string temp_string;
        _add_symbol(temp_string, (char*)" ", _line_width - message_lenght - 1); 
        out << temp_string << "|" << std::endl;
        
        temp_string.clear();
        _add_symbol(temp_string, (char*)" ", first_column_width); 
        out << "| " << temp_string << " | " << std::dec << failure.value << " time(s)";
        temp_string.clear();
        _add_symbol(temp_string, (char*)" ", _line_width - 2- first_column_width - 3 - _get_size(failure.value) - ((std::string)" time(s)").size() - 1); 
        out << temp_string << "|" << std::endl;
    }
    else
    {
        // Print failure amount
        out << " | " << std::dec << failure.value << " time(s)";

        message_lenght += times_lenght - 2; // minus " |".size()
        
        // Add spaces befor print end of line ("|")
        std::string temp_string;
        _add_symbol(temp_string, (char*)" ", _line_width - message_lenght - 1); 
        out << temp_string << "|" << std::endl;
    }
    
    // Print separation line like this: "|----------|--------|---------|"
    std::string separation_line;
    separation_line.append("|");
    int column_width = 1 + first_column_width + 1;
    _add_symbol(separation_line, (char*)"-", column_width);
    separation_line.append("|");
    _add_symbol(separation_line, (char*)"-", _line_width - column_width - 3);
    separation_line.append("|");
    
    out << separation_line << std::endl;
}

/// Print statistic part of report
void ConsolePrinter::_print_statistic(std::ostream &out, const cpptesk::hw::report::Statistics &statistics)
{
    // Print statistic title
    out << "| ";
    
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(statistics.failure ? RED : GREEN, NORMAL); 
    out << statistics.name;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    out << " | ";

    out << statistics.stimuli.name << " | ";
    out << statistics.reactions.name << " | ";
    out << statistics.normal.name << " | ";

    if(_debug_color) out << CPPTESK_DEBUG_STYLE(statistics.incorrect.failure ? RED : GREEN, NORMAL); 
    out << statistics.incorrect.name;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    out << " | ";

    bool long_missing = false;
    if((statistics.expected_missing.value != 0 || statistics.optional_missing.value != 0) && statistics.missing.value > 0)
    {
        std::stringstream temp;
        
        temp << statistics.expected_missing.value + statistics.optional_missing.value << "+" << statistics.missing.value;
        std::string temp_missing_str = temp.str();

        if(statistics.missing.name.size() < temp_missing_str.size())
            long_missing = true;
    }
    
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(statistics.missing.failure ? RED : GREEN, NORMAL); 
    out << statistics.missing.name;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    out << " | ";

    if(_debug_color) out << CPPTESK_DEBUG_STYLE(statistics.unexpected.failure ? RED : GREEN, NORMAL); 
    out << statistics.unexpected.name;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    
    out << " | "<< std::endl;
    
    // Print separation line like this: "|----------|--------|---------|"
    std::string separation_line;
    separation_line.append("|");
    _add_symbol(separation_line, (char*)"-", statistics.name.size() + 2); // 2 = 2 spaces, left+right of "_NAME_"
    separation_line.append("|");
    _add_symbol(separation_line, (char*)"-", statistics.stimuli.name.size() + 2);
    separation_line.append("|");
    _add_symbol(separation_line, (char*)"-", statistics.reactions.name.size() + 2);
    separation_line.append("|");
    _add_symbol(separation_line, (char*)"-", statistics.normal.name.size() + 2);
    separation_line.append("|");
    _add_symbol(separation_line, (char*)"-", statistics.incorrect.name.size() + 2);
    separation_line.append("|");
    _add_symbol(separation_line, (char*)"-", statistics.missing.name.size() + 2);
    separation_line.append("|");
    _add_symbol(separation_line, (char*)"-", statistics.unexpected.name.size() + 2);
    separation_line.append("|");
    out << separation_line << std::endl;
    
    // Print statistic value
    int statistics_precision = 2;
    if(statistics.value > 10.0)  statistics_precision = 1;
    if(statistics.value > 100.0) statistics_precision = 0;
    
    out << "| " << std::fixed << std::setprecision(statistics_precision) << std::dec 
                           << statistics.value << " " << statistics.unit << " | ";
                           
    std::string temp_str;
    int temp_int = statistics.stimuli.name.size() - _get_size(statistics.stimuli.value);
    _add_symbol(temp_str, (char*)" ", temp_int > 0 ? temp_int : 0);
    out << temp_str << std::dec << statistics.stimuli.value << " | ";

    temp_str.clear();
    temp_int = statistics.reactions.name.size() - _get_size(statistics.reactions.value);
    _add_symbol(temp_str, (char*)" ", temp_int > 0 ? temp_int : 0);
    out << temp_str << std::dec << statistics.reactions.value << " | ";
    
    temp_str.clear();
    temp_int = statistics.normal.name.size() - _get_size(statistics.normal.value);
    _add_symbol(temp_str, (char*)" ", temp_int > 0 ? temp_int : 0);
    out << temp_str << std::dec << statistics.normal.value << " | ";

    temp_str.clear();
    temp_int = statistics.incorrect.name.size() - _get_size(statistics.incorrect.value);
    _add_symbol(temp_str, (char*)" ", temp_int > 0 ? temp_int : 0);
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(statistics.incorrect.failure ? RED : GREEN, NORMAL); 
    out << temp_str << std::dec << statistics.incorrect.value;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    out << " | ";

    if(statistics.missing.value > 0)
    {
        if(statistics.expected_missing.value == 0 &&  statistics.optional_missing.value == 0)
        {
            temp_str.clear();
            temp_int = statistics.missing.name.size() - _get_size(statistics.missing.value);
            _add_symbol(temp_str, (char*)" ", temp_int > 0 ? temp_int : 0);
            if(_debug_color) out << CPPTESK_DEBUG_STYLE(RED, NORMAL); 
            out << temp_str << std::dec << statistics.missing.value;
        }
        else
        {
            temp_str.clear();
            
            if(!long_missing)
            {
                temp_int = statistics.missing.name.size() - _get_size(statistics.missing.value) 
                        - _get_size(statistics.expected_missing.value + statistics.optional_missing.value) - 1; // 1 = "+".size()
                _add_symbol(temp_str, (char*)" ", temp_int > 0 ? temp_int : 0);
                if(_debug_color) out << CPPTESK_DEBUG_STYLE(RED, NORMAL); 
                out << temp_str << std::dec << statistics.missing.value;
                if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
                out << "+";
                if(_debug_color) out << CPPTESK_DEBUG_STYLE(GREEN, NORMAL);
                out << std::dec << statistics.expected_missing.value + statistics.optional_missing.value; 
            }
            else
            {
                temp_int = statistics.missing.name.size() - _get_size(statistics.missing.value) - 1;
                _add_symbol(temp_str, (char*)" ", temp_int > 0 ? temp_int : 0);
                if(_debug_color) out << CPPTESK_DEBUG_STYLE(RED, NORMAL); 
                out << temp_str << std::dec << statistics.missing.value;
                if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
                out << "+";
            }
        }
    }
    else
    {
        temp_str.clear();
        temp_int = statistics.missing.name.size() - _get_size(statistics.expected_missing.value + statistics.optional_missing.value);
        _add_symbol(temp_str, (char*)" ", temp_int > 0 ? temp_int : 0);
        if(_debug_color) out << CPPTESK_DEBUG_STYLE(GREEN, NORMAL); 
        out << temp_str << std::dec << statistics.expected_missing.value + statistics.optional_missing.value; 
    }
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    out << " | ";

    temp_str.clear();
    temp_int = statistics.unexpected.name.size() - _get_size(statistics.unexpected.value);
    _add_symbol(temp_str, (char*)" ", temp_int >0 ? temp_int : 0);
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(statistics.unexpected.failure ? RED : GREEN, NORMAL); 
    out << temp_str << std::dec << statistics.unexpected.value;
    if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    out << " |";

    out << std::endl;
 
    // Additional line for results 
    if(long_missing)
    {
        separation_line.clear();
        separation_line.append("|");
        _add_symbol(separation_line, (char*)" ", statistics.name.size() + 2); // 2 = 2 spaces, left+right of "_NAME_"
        separation_line.append("|");
        _add_symbol(separation_line, (char*)" ", statistics.stimuli.name.size() + 2);
        separation_line.append("|");
        _add_symbol(separation_line, (char*)" ", statistics.reactions.name.size() + 2);
        separation_line.append("|");
        _add_symbol(separation_line, (char*)" ", statistics.normal.name.size() + 2);
        separation_line.append("|");
        _add_symbol(separation_line, (char*)" ", statistics.incorrect.name.size() + 2);
        separation_line.append("| ");
        out << separation_line;
        
        temp_str.clear();
        temp_int = statistics.missing.name.size() - _get_size(statistics.expected_missing.value + statistics.optional_missing.value);
        _add_symbol(temp_str, (char*)" ", temp_int > 0 ? temp_int : 0);
        out << temp_str;
        
        if(_debug_color) out << CPPTESK_DEBUG_STYLE(GREEN, NORMAL);
        out << std::dec << statistics.expected_missing.value + statistics.optional_missing.value; 
        if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);

        separation_line.clear();
        separation_line.append(" |");
        _add_symbol(separation_line, (char*)" ", statistics.unexpected.name.size() + 2);
        separation_line.append("|");
        out << separation_line << std::endl;
    }
    
    _print_line(out, _line_width); // Print separation line
}

/// Print simulation part of report
void ConsolePrinter::_print_simulation(std::ostream &out, const cpptesk::hw::report::Simulation &simulation)
{
    int message_lenght = 0; // The width of output string
    
    out << "| " << simulation.name << " | ";
    
    // Print simulation information
    out << std::dec << simulation.cycle.value << " " << simulation.cycle.unit << " / ";
    out << std::dec << simulation.time.value << " " << simulation.time.unit << " / ";
    out << std::fixed << std::setprecision(_frequency_precision) << std::dec << 
                              simulation.frequency.value << " " << simulation.frequency.unit;
    
    // The width of output string
    message_lenght = 2 + ((std::string)simulation.name).size() + 3;
    message_lenght += _get_size(simulation.cycle.value) + 1 + ((std::string)simulation.cycle.unit).size() + 3;
    message_lenght += _get_size(simulation.time.value) + 1 + ((std::string)simulation.time.unit).size() + 3;
    message_lenght += _get_size(simulation.frequency.value, _frequency_precision) + 1 + ((std::string)simulation.frequency.unit).size();
        
    // Add spaces befor print end of line ("|")
    std::string temp_string;
    _add_symbol(temp_string, (char*)" ", _line_width - message_lenght - 1); 
    out << temp_string << "|" << std::endl;
    
    _print_line(out); // Print separation line
}

/// Print explanation only for spec reaction
void ConsolePrinter::_print_explanation(std::ostream &out, std::string explanation)
{
    std::string separation_line;
    separation_line.append("|");
    _add_symbol(separation_line, (char*)"-", _line_width - 2);
    separation_line.append("|");

    out << separation_line << std::endl;
    
    const unsigned DOUBLE_LENGHT_SHIFT = 4; // "| " & "| "
    const unsigned LENGHT_SHIFT = 2; // "| " or "| "
    const unsigned NEXT_LINE_PART = 3;

    std::string str_explanation;
    
    unsigned point = 0;
    
    unsigned found_point = explanation.find("\n");
    if((found_point != std::string::npos) && (found_point == 0))
        point++;

    while(true)
    {
        if(explanation.length() - point <= _line_width - DOUBLE_LENGHT_SHIFT)
        {
            explanation = explanation.substr(point, explanation.length());
            str_explanation = "| ";
            str_explanation.append(explanation);
            _add_symbol(str_explanation, (char*)" ", (_line_width - str_explanation.length() - LENGHT_SHIFT));
            str_explanation.append(" |");
            
            out << str_explanation << std::endl;
                    
            break;
        }

        str_explanation = explanation.substr(point, _line_width - DOUBLE_LENGHT_SHIFT + 1);
        
        found_point = str_explanation.find_last_of(" ");
        
        if(found_point == _line_width - DOUBLE_LENGHT_SHIFT) // If " " the last+1 symbol of line, then ignore him.
        {
            found_point--;
            point++;
        }   
        
        if((found_point < (NEXT_LINE_PART - 1)*_line_width/NEXT_LINE_PART) or (found_point == std::string::npos))
        {
            found_point = str_explanation.find_last_of(",");
        
            if((found_point < (NEXT_LINE_PART - 1)*_line_width/NEXT_LINE_PART) or (found_point == std::string::npos))
                found_point = _line_width - DOUBLE_LENGHT_SHIFT - 1;
        }
            
        point = point + found_point + 1;
        
        str_explanation = str_explanation.substr(0, found_point + 1);

        str_explanation.insert(0, "| ");
        _add_symbol(str_explanation, (char*)" ", (_line_width - str_explanation.length() - LENGHT_SHIFT));
        str_explanation.append(" |");
        
        out << str_explanation << std::endl;
    }
}

void ConsolePrinter::_print_color_field(std::ostream &out, const cpptesk::hw::report::FieldList &field_list, const debug_color_t color)
{
    for(cpptesk::hw::report::FieldList::const_iterator i = field_list.begin(); i != field_list.end(); i++)
    {
        const cpptesk::hw::report::ItemField &item = *i;

        if(_debug_color) out << CPPTESK_DEBUG_STYLE(item.failure ? color : NORMAL, NORMAL);
        out << item.value;
        if(_debug_color) out << CPPTESK_DEBUG_STYLE(NORMAL, NORMAL);
    }                
}

/// Return size of string of digital value
int ConsolePrinter::_get_size(int temp_num)
{
    std::stringstream temp;
    temp << temp_num;
    return temp.str().size();
}

int ConsolePrinter::_get_size(double temp_num, unsigned precision)
{
    std::stringstream temp;
    temp << std::fixed << std::setprecision(precision) << std::dec << temp_num;
    return temp.str().size();
}

/// Print separation line
void ConsolePrinter::_print_line(std::ostream &out, int length)
{
    std::string str_big_line;
    _add_symbol(str_big_line, (char*)"=", length);
    out << str_big_line << std::endl;
}

void ConsolePrinter::_print_line(std::ostream &out)
{
    _print_line(out, _line_width);
}

/// Add symbol to string
inline void ConsolePrinter::_add_symbol_top(std::string &str, char* symbol, const int iterator)
{
    for(int i = 0; i < iterator; i++)
        str.insert(0, symbol);
}

inline void ConsolePrinter::_add_symbol(std::string &str, char* symbol, const int iterator)
{
    for(int i = 0; i < iterator; i++)
        str.append(symbol);
}

/// Return max width of strings
int ConsolePrinter::get_max_width(const std::string &str_0, const std::string &str_1, const std::string &str_2)
{
    if(str_0.size() > str_1.size() && str_0.size() > str_2.size())
    {
        return str_0.size();
    }
    else if(str_1.size() > str_2.size())
    {
        return str_1.size();    
    }
    else return  str_2.size();    
}

}}} 
