/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Report model
// ///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>
#include <hw/time.hpp>
#include <hw/utils/range.hpp>

#include <list>
#include <string>

using namespace cpptesk::hw;

namespace cpptesk {
namespace hw {
namespace report {

//------------------------------------------------------------------------------
// Report Ingredients
//------------------------------------------------------------------------------

/**
 * This structure represents an entity.
 */
template <typename T>
struct Entity
{
    /// The entity name or identifier.
    std::string name;
    /// The entity type.
    std::string type;
    /// The entity value.
    T value;
    /// The unit of measurement.
    std::string unit;
    /// The entity description.
    std::string info;
    /// The failure indicator.
    bool failure;

    /// Constructs an entity with given values of the attributes.
    Entity(const std::string &name, const std::string &type, const T &value, bool failure = false):
        name(name), type(type), value(value), failure(failure) {}
    /// Constructs an entity with the default value.
    Entity(const std::string &name, const std::string &type, bool failure = false):
        name(name), type(type), failure(failure) {}
    /// Constructs an entity with the default values of the attributes.
    Entity():
        name("<name>"), type("<type>"), failure(false) {}
    /// Destructs the entity.
    virtual ~Entity() {}
};

/**
 * This structure represents a matching (a pair of entities of the same type).
 */
template <typename T>
struct Matching: public Entity<T>
{
    /// The specification value.
    T spec;
    /// The implementation value.
    T impl;
    
    /// Constructs a matching with given values of the attributes.
    Matching(const std::string &name, const std::string &type, const T &spec, const T &impl, bool failure = false):
        Entity<T>(name, type, spec, failure), spec(spec), impl(impl) {}
    /// Constructs a matching with the default value.
    Matching(const std::string &name, const std::string &type, bool failure = false):
        Entity<T>(name, type, failure) {}
    /// Constructs a matching with the default values of the attributes.
    Matching():
        Entity<T>() {}
    /// Destructs the matching.
    virtual ~Matching() {}
};

/**
 * This structure implements an entity with list-type values.
 */
template <typename T>
struct List: public Entity< std::list<T> >
{
    /// The iterator type.
    typedef typename std::list<T>::iterator iterator;
    /// The constant iterator type.
    typedef typename std::list<T>::const_iterator const_iterator;

    /// Constructs a list with given attributes.
    List(const std::string &name, const std::string &type, bool failure = false):
        Entity< std::list<T> >(name, type, failure) {}
    /// Constructs a list with default values of attributes.
    List() {}
    /// Destructs the list.    
    virtual ~List() {}
    
    /// Returns the iterator pointing to the beginning of the list.
    iterator begin() { return this->value.begin(); }
    /// Returns the end value of the iterator.
    iterator end() { return this->value.end(); }

    /// Returns the constant iterator pointing to the beginning of the list.
    const_iterator begin() const { return this->value.begin(); }
    /// Returns the end value of the constant iterator.
    const_iterator end() const { return this->value.end(); }

    /// Appends a given item to the list.
    void add(const T &item) { this->value.push_back(item); }
};

/**
 * This structure describes field items.
 */
struct ItemField: public Entity<std::string>  
{
    void setFailure() {this->failure = true;}
};

/**
 * This structure contain items
 */
struct FieldList: public List<ItemField> 
{
    /// Return field size
    int size() {return this->value.size();}
    
    std::string toString() const
    {
        std::string temp;
        for(const_iterator i = begin(); i != end(); i++)
        {
            const ItemField &item = *i;
            temp.append(item.value);
        }
        return temp;
    }
};

/**
 * This structure describes field failures.
 */
struct FieldFailure: public Matching<FieldList>
{
    /// The comparable indicator.
    bool comparable;
};

/**
 * This structure describes data failures.
 */
struct DataFailure: public List<FieldFailure> {};

/**
 * This structure describes time failures.
 */
struct TimeFailure: public Matching<std::string> {};

/**
 * This structure describes interface failures.
 */
struct IfaceFailure: public Matching<std::string> {};

/**
 * This structure describes a failure.
 *
 * The failure fields have the following meaning: 
 *   - failure.name    - the identifier (usually, #1, #2, #3, etc.);
 *   - failure.type    - the type (INCORRECT, MISSING, UNEXPECTED, etc.);
 *   - failure.value   - the multiplicity (a possitive integer value);
 *   - failure.units   - the unit for measuring the multiplicity (usually, times);
 *   - failure.info    - the description (a user-defined text);
 *   - failure.failure - the indicator (usually, true);
 *  
 */
struct Failure: public Entity<unsigned>
{
    /// The data failure.
    DataFailure  data;
    /// The timing failure.
    TimeFailure  time;
    /// The interface failure.
    IfaceFailure iface;
};

/**
 * This structure contain list of identical failures.
 */
struct FailureGroup: public List<Failure> {};

/**
 * This structure represents a count of events.
 */
struct Count: public Entity<unsigned> {};

/**
 * This structure represents information about failures one type.
 */
struct Failures: public List<FailureGroup> {};

/**
 * This structure represents simulation time information.
 */
struct Simulation: public Entity<unsigned> 
{
    /// The cycles number
    Entity<cycle_t> cycle;
    /// The simulation time
    Entity<time_t> time;
    /// The frequency of simulation 
    Entity<double> frequency;
};

/**
 * This structure represents statistics on events.
 */
struct Statistics: public Entity<double> 
{
    /// The information about stimuli
    Entity<unsigned> stimuli;
    /// The information about reactions
    Entity<unsigned> reactions;
    
    /// The information about normal reactions
    Entity<unsigned> normal;
    /// The information about incorrect reactions
    Entity<unsigned> incorrect;
    /// The information about missing reactions
    Entity<unsigned> missing;
    /// The information about not error missing reactions
    Entity<unsigned> expected_missing;
    Entity<unsigned> optional_missing;

    /// The information about unexpected reactions
    Entity<unsigned> unexpected;
};

/**
 * This structure represents a report.
 */
struct Report
{
    /// The information about failures.
    Failures failures;
    /// The simulation time information.
    Simulation simulation;
    /// The stimuli/reactions statistics.
    Statistics statistics;
};

}}} // namespace cpptesk::hw::report

