/*
 * Copyright 2010-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Protsenko (protsenko@ispras.ru)
// SYNOPSIS: Class for Printers of Diagnostics Results in xml format
///////////////////////////////////////////////////////////////////////////////

#include <hw/interface.hpp>
#include <hw/report/xml_printer.hpp>

namespace cpptesk {
namespace hw {
namespace report {

std::string Tag::toString()
{
    std::stringstream stream;
    
    stream << "<" << name;
    
    for(uint i = 0; i < params.size(); i++)
    {
        stream << " " << params[i].name << "=" << params[i].value;
    }
    
    stream << ">" << std::endl;
    
    for(uint i = 0; i < tags.size(); i++)
    {
        Tag &subtag = tags[i];
        stream << subtag.toString();
    }
    
    stream << "</" << name << ">" << std::endl;
    
    return stream.str();
}

std::string Tag::operator<<(Tag &tag)
{
    return tag.toString();
}

XMLPrinter::XMLPrinter()
{
    _xml_trace = new std::ofstream();
    _tab = "    ";
}

XMLPrinter::~XMLPrinter()
{
    delete _xml_trace;
}

void XMLPrinter::print(std::string xml_trace_name, const Report &report)
{
    _xml_trace->open(xml_trace_name.c_str(), std::fstream::out | std::fstream::trunc);
    print(*_xml_trace, report);
    _xml_trace->close();
}

void XMLPrinter::print(std::ostream &out, const Report &report)
{
    out << "<!-- Diagnostics Results -->" << std::endl << std::endl;
    
    Tag failures_tag = Tag("failures_list");
    
    _process_group(failures_tag, report.failures);

    Tag statistics_tag = Tag("statistics");
    
    _process_statistics(statistics_tag, report.statistics);
    
    Tag simulation_tag = Tag("simulation");
    
    _process_simulation(simulation_tag, report.simulation);
    
    Tag report_tag = Tag("report");
    
    report_tag.add_subtag(simulation_tag);
    report_tag.add_subtag(statistics_tag);
    report_tag.add_subtag(failures_tag);
    
    _print_tag(report_tag, 0);
}

Tag XMLPrinter::_process_simulation(Tag &simulation_tag, const cpptesk::hw::report::Simulation &simulation)
{
    simulation_tag.add_parameter(TagParameter("name", simulation.name));    
    
    Tag cycle_tag = Tag("cycle");
    cycle_tag.add_parameter(TagParameter("value", simulation.cycle.value));    
    cycle_tag.add_parameter(TagParameter("unit", simulation.cycle.unit));    
    
    Tag time_tag = Tag("time");
    time_tag.add_parameter(TagParameter("value", simulation.time.value));    
    time_tag.add_parameter(TagParameter("unit", simulation.time.unit));    

    Tag frequency_tag = Tag("frequency");
    frequency_tag.add_parameter(TagParameter("value", simulation.frequency.value));    
    frequency_tag.add_parameter(TagParameter("unit", simulation.frequency.unit));    

    simulation_tag.add_subtag(cycle_tag);
    simulation_tag.add_subtag(time_tag);
    simulation_tag.add_subtag(frequency_tag);
    
    return simulation_tag;
}

Tag XMLPrinter::_process_statistics(Tag &statistics_tag, const cpptesk::hw::report::Statistics &statistics)
{
    statistics_tag.add_parameter(TagParameter("failure", statistics.failure));
    statistics_tag.add_parameter(TagParameter("name", statistics.name));    
    
    Tag stimuli_tag = Tag("stimuli");
    stimuli_tag.add_parameter(TagParameter("name", statistics.stimuli.name));    
    stimuli_tag.add_parameter(TagParameter("value", statistics.stimuli.value));
    
    Tag reactions_tag = Tag("reactions");
    reactions_tag.add_parameter(TagParameter("name", statistics.reactions.name));    
    reactions_tag.add_parameter(TagParameter("value", statistics.reactions.value));

    Tag normal_tag = Tag("normal");
    normal_tag.add_parameter(TagParameter("name", statistics.normal.name));    
    normal_tag.add_parameter(TagParameter("value", statistics.normal.value));

    Tag incorrect_tag = Tag("incorrect");
    incorrect_tag.add_parameter(TagParameter("name", statistics.incorrect.name));    
    incorrect_tag.add_parameter(TagParameter("value", statistics.incorrect.value));
    incorrect_tag.add_parameter(TagParameter("failure", statistics.incorrect.failure));

    Tag expected_missing_tag = Tag("expected_missing");
    expected_missing_tag.add_parameter(TagParameter("name", statistics.expected_missing.name));    
    expected_missing_tag.add_parameter(TagParameter("value", statistics.expected_missing.value));

    Tag missing_tag = Tag("missing");
    missing_tag.add_parameter(TagParameter("name", statistics.missing.name));    
    missing_tag.add_parameter(TagParameter("value", statistics.missing.value));
    missing_tag.add_parameter(TagParameter("failure", statistics.missing.failure));

    Tag unexpected_tag = Tag("unexpected");
    unexpected_tag.add_parameter(TagParameter("name", statistics.unexpected.name));    
    unexpected_tag.add_parameter(TagParameter("value", statistics.unexpected.value));
    unexpected_tag.add_parameter(TagParameter("failure", statistics.unexpected.failure));

    statistics_tag.add_subtag(stimuli_tag);
    statistics_tag.add_subtag(reactions_tag);
    statistics_tag.add_subtag(normal_tag);
    statistics_tag.add_subtag(incorrect_tag);
    statistics_tag.add_subtag(expected_missing_tag);
    statistics_tag.add_subtag(missing_tag);
    statistics_tag.add_subtag(unexpected_tag);
    
    return statistics_tag;
}
    
Tag XMLPrinter::_process_group(Tag &failures_tag, const cpptesk::hw::report::Failures &failures)
{
    for(cpptesk::hw::report::Failures::const_iterator i = failures.begin(); i != failures.end(); i++)
    {
        const cpptesk::hw::report::FailureGroup &failure_group = *i;
    
        Tag group_tag = Tag("failure_group");
        
        const cpptesk::hw::report::Failure &failure = *failure_group.begin();
        
        group_tag.add_parameter(TagParameter("number", (std::string)failure.name));
        group_tag.add_parameter(TagParameter("value", (uint)failure.value));
        group_tag.add_parameter(TagParameter("type", (std::string)failure.type));
        group_tag.add_parameter(TagParameter("iface", (std::string)failure.iface.name));
        group_tag.add_parameter(TagParameter("time", (std::string)failure.time.value));

        group_tag.add_subtag(_process_failure(failure));
        
        failures_tag.add_subtag(group_tag);
    }

    return failures_tag;
}

void XMLPrinter::_print(std::string str, uint tab)
{
    for(uint i = 0; i < tab; i++)
        (*_xml_trace) << _tab;
    (*_xml_trace) << str;

    (*_xml_trace) << std::endl;
}

void XMLPrinter::_print_tag_border(std::string tag_name, bool left, uint tab)
{
    std::stringstream stream;
    stream << "<" << (left ? "" : "/") << tag_name << ">";
    _print(stream.str(), tab);
}

void XMLPrinter::_print_open_tag(std::string tag_name, uint tab)
{
    _print_tag_border(tag_name, true, tab);
}

void XMLPrinter::_print_close_tag(std::string tag_name, uint tab)
{
    _print_tag_border(tag_name, false, tab);
}

void XMLPrinter::_print_param(TagParameter param, uint tab)
{
    _print_open_tag(param.name, tab + 1);
    _print(param.value, tab + 1 + 1);
    _print_close_tag(param.name, tab + 1);
}

void XMLPrinter::_print_tag(Tag tag, uint tab)
{
    _print_open_tag(tag.name, tab);
    for(uint i = 0; i < tag.params.size(); i++)
        _print_param(tag.params[i], tab);
    for(uint i = 0; i < tag.tags.size(); i++)
        _print_tag(tag.tags[i], tab + 1);
    _print_close_tag(tag.name, tab);
}

Tag XMLPrinter::_process_failure(const cpptesk::hw::report::Failure &failure)
{
    Tag failure_tag("failure");
    
    failure_tag.add_subtag(_process_data(failure.data));
    failure_tag.add_subtag(_process_time(failure.time));
    failure_tag.add_subtag(_process_iface(failure.iface));
        
    return failure_tag;
}

Tag XMLPrinter::_process_data(const cpptesk::hw::report::DataFailure &data)
{
    Tag data_tag("data");
    
    data_tag.add_parameter(TagParameter("failure", data.failure));
    
    for(cpptesk::hw::report::DataFailure::const_iterator i = data.begin(); i != data.end(); i++)
    {
        const cpptesk::hw::report::FieldFailure &field = *i;

        Tag field_tag("field");
        field_tag.add_parameter(TagParameter("name", field.name));
        
        field_tag.add_subtag(_process_field(field.spec, "spec"));
        field_tag.add_subtag(_process_field(field.impl, "impl"));
        
        data_tag.add_subtag(field_tag);
    }

    return data_tag;
}

Tag XMLPrinter::_process_field(const cpptesk::hw::report::FieldList &field_list, const std::string name)
{
    Tag field_list_tag(name);
    
    for(cpptesk::hw::report::FieldList::const_iterator i = field_list.begin(); i != field_list.end(); i++)
    {
        const cpptesk::hw::report::ItemField &item = *i;

        Tag item_tag("item");
        item_tag.add_parameter(TagParameter("value", item.value));
        item_tag.add_parameter(TagParameter("failure", item.failure));
        
        field_list_tag.add_subtag(item_tag);
    }                
    
    return field_list_tag;
}

Tag XMLPrinter::_process_time(const cpptesk::hw::report::TimeFailure &time)
{
    Tag time_tag("time");
    time_tag.add_parameter(TagParameter("spec", time.spec));
    return time_tag;
}

Tag XMLPrinter::_process_iface(const cpptesk::hw::report::IfaceFailure &iface)
{
    Tag iface_tag("iface");
    iface_tag.add_parameter(TagParameter("spec", iface.spec));
    return iface_tag;
}
}}} // cpptesk::hw::report
