/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Hash table
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <list>
#include <map>
#include <vector>

#include <hw/defs.hpp>

namespace cpptesk {
namespace hw {

template <typename K, typename V>
class HashTable
{
private:
    typedef std::list<V>                              _items_t;
    typedef std::pair<K, typename _items_t::iterator> _entry_t;
    typedef std::list<_entry_t>                       _variants_t;
    typedef std::vector<_variants_t>                  _table_t;

public:
    typedef typename _items_t::iterator iterator;

    HashTable(size_t table_size):
        _all_items(), _table(table_size), _table_size(table_size), _size(0) {}

    ~HashTable() {}

    /// IMPORTANT: This method should be called before filling the hash table.
    void resize(size_t new_size)
    {
        _table.resize(_table_size = new_size);
    }

    void put(K key, V value)
    {
        _variants_t &variants = access(key);

        for(typename _variants_t::iterator i = variants.begin(); i != variants.end(); i++)
        {
            if(i->first == key)
                { return; }
        }

        _all_items.push_back(value);
        variants.push_back(_entry_t(key, (++_all_items.rbegin()).base()));

        _size++;
    }

    void remove(K key)
    {
        _variants_t &variants = access(key);

        for(typename _variants_t::iterator i = variants.begin(); i != variants.end(); i++)
        {
            if(i->first == key)
            {
                _all_items.erase(i->second);
                variants.erase(i);

                _size--;

                break;
            }
        }
    }

    iterator find(K key)
    {
        _variants_t &variants = access(key);

        for(typename _variants_t::iterator i = variants.begin(); i != variants.end(); i++)
        {
            if(i->first == key)
                { return i->second; }
        }

        return end();
    }

    size_t size() const { return _size; }

    iterator begin() { return _all_items.begin(); }
    iterator end()   { return _all_items.end(); }
    
private:
    _variants_t& access(K key) { return _table[key % _table_size]; }
    
    _items_t  _all_items;
    _table_t  _table;
    size_t    _table_size;
    size_t    _size;
};

}} // namespace cpptesk::hw
