/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Multicore cache-coherent processor.
///////////////////////////////////////////////////////////////////////////////

#include <hw/assertion.hpp>
#include <moesi/processor.hpp>

namespace cpptesk {
namespace examples {
namespace moesi {

///////////////////////////////////////////////////////////////////////////////
// Processor core
///////////////////////////////////////////////////////////////////////////////
Core::Core():
    _id(0),
    _processor(0),
    _protocol(0)
{}

Core::~Core()
{
    if(_protocol)
        { delete _protocol; }
}

Processor& Core::get_processor()
{
    CPPTESK_ASSERTION(_processor, "The parent processor has not been set");
    return *_processor;
}

const Processor& Core::get_processor() const
{
    CPPTESK_ASSERTION(_processor, "The parent processor has not been set");
    return *_processor;
}

void Core::set_processor(Processor *processor)
{
    CPPTESK_ASSERTION(processor, "The parent processor is null");
    _processor = processor;
}

void Core::set_protocol(Protocol *protocol)
{
    CPPTESK_ASSERTION(protocol, "The cache coherence protocol is null");
    (_protocol = protocol)->start();
}

size_t Core::get_cores_number() const
{
    CPPTESK_ASSERTION(_processor, "The parent processor has not been set");
    return _processor->get_cores_number();
}

Core& Core::get_core(size_t i)
{
    CPPTESK_ASSERTION(_processor, "The parent processor has not been set");
    return _processor->get_core(i);
}

const Core& Core::get_core(size_t i) const
{
    CPPTESK_ASSERTION(_processor, "The parent processor has not been set");
    return _processor->get_core(i);
}

bool Core::pre_recv(CacheStimulus stimulus) const
{
    CPPTESK_ASSERTION(_protocol, "The cache coherence protocol has not been set");
    return _protocol->precondition(stimulus);
}

void Core::recv(CacheStimulus stimulus)
{
    CPPTESK_ASSERTION(_protocol, "The cache coherence protocol has not been set");
    _protocol->step(stimulus);
}

void Core::send(CacheStimulus stimulus, size_t i)
{
    CPPTESK_ASSERTION(_processor, "The parent processor has not been set");
    get_core(i).recv(stimulus);
}

void Core::broadcast(CacheStimulus stimulus)
{
    CPPTESK_ASSERTION(_processor, "The parent processor has not been set");

    for(size_t i = 0; i < _processor->get_cores_number(); i++)
    {
        if(i != get_id())
            { send(stimulus, i); }
    }
}

CacheState Core::get_state() const
{
    CPPTESK_ASSERTION(_protocol, "The cache coherence protocol has not been set");
    return CacheState(_protocol->state());
}

void Core::reset()
{
    CPPTESK_ASSERTION(_protocol, "The cache coherence protocol has not been set");
    _protocol->stop();
    _protocol->start();
}

///////////////////////////////////////////////////////////////////////////////
// Multi-core processor
///////////////////////////////////////////////////////////////////////////////

Processor::Processor(CacheProtocol protocol, size_t n):
    _cores(n)
{
    for(cores_t::iterator i = _cores.begin(); i != _cores.end(); i++)
    {
        i->set_id(i - _cores.begin());
        i->set_processor(this);
	i->set_protocol(Protocol::create(protocol, *i));
    }
}

Processor::~Processor()
{}

bool Processor::pre_recv(CacheStimulus stimulus, size_t i) const
{
    return _cores[i].pre_recv(stimulus);
}

void Processor::recv(CacheStimulus stimulus, size_t i)
{
    _cores[i].recv(stimulus);
}

bool Processor::is_initial_state() const
{
    for(cores_t::const_iterator i = _cores.begin(); i != _cores.end(); i++)
    {
        if(i->get_state() != INVALID)
            { return false; }
    }

    return true;
}

void Processor::reset()
{
    for(cores_t::iterator i = _cores.begin(); i != _cores.end(); i++)
        { i->reset(); }
}

}}} // cpptesk::examples::moesi
