/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Multicore cache-coherent processor.
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <moesi/protocol.hpp>

#include <sstream>
#include <vector>

namespace cpptesk {
namespace examples {
namespace moesi {

////////////////////////////////////////////////////////////////////////////////
// Core and multi-core processor
////////////////////////////////////////////////////////////////////////////////
class Processor;

/**
 * Processor core.
 */
class Core
{
    friend class Processor;

public:
    /// Core constructor.
    Core();
    /// Core destructor.
   ~Core();

    /// Returns the core identifier.
    size_t get_id() const { return _id; }

    /// Returns the parent multi-core processor.
    Processor& get_processor();
    /// Returns the parent multi-core processor (constant reference).
    const Processor& get_processor() const;

    /// Returns the overall number of cores.
    size_t get_cores_number() const;

    /// Returns the i-th core of the parent processor.
    Core& get_core(size_t i);
    /// Returns the i-th core of the parent processor (constant reference).
    const Core& get_core(size_t i) const;

    /// Checks whether the stimulus is acceptable.
    bool pre_recv(CacheStimulus stimulus) const;

    /// Receives the message.
    void recv(CacheStimulus stimulus);
    /// Sends the message.
    void send(CacheStimulus stimulus, size_t i);
    /// Broadcasts the message.
    void broadcast(CacheStimulus stimulus);

    /// Returns the cache line state.
    CacheState get_state() const;

    /// Resets the cache line state.
    void reset();

private:
    /// Sets the core identifier.
    void set_id(size_t id) { _id = id; }
    /// Sets the parent multi-core processor.
    void set_processor(Processor *processor);
    /// Sets the cache coherence protocol.
    void set_protocol(Protocol *protocol);

private:
    /// Core identifier.
    size_t _id;
    /// Parent multi-core processor.
    Processor *_processor;
    /// Cache coherence protocol.
    Protocol *_protocol;
};

/**
 * Multi-core processor.
 */
class Processor
{
public:
    /// Processor constructor.
    Processor(CacheProtocol protocol, size_t n);
    /// Processor destructor.
   ~Processor();

    /// Retursn the cache coherence protocol.
    CacheProtocol get_protocol() const { return _protocol; }

    /// Returns the number of cores.
    size_t get_cores_number() const { return _cores.size(); }

    /// Returns the i-th core.
    Core& get_core(size_t i) { return _cores[i]; }
    /// Returns the i-th core (constant reference).
    const Core& get_core(size_t i) const { return _cores[i]; }

    /// Checks whether the stimulus is acceptable.
    bool pre_recv(CacheStimulus stimulus, size_t i) const;
    /// Executes the operation.
    void recv(CacheStimulus stimulus, size_t i);

    /// Checks whether the state is initial (invalid).
    bool is_initial_state() const;

    /// Resets the state.
    void reset();

private:
    /// Processor cores type.
    typedef std::vector<Core> cores_t;
    /// Cache coherence protocol.
    CacheProtocol _protocol;
    /// Processor cores.
    cores_t _cores;
};

}}} // cpptesk::examples::moesi

