/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Family of models for the MOESI cache coherence protocols:
//           (1) MSI   (Modified,                   Shared, Invalid)
//           (2) MOSI  (Modified, Owned,            Shared, Invalid)
//           (3) MESI  (Modified,        Exclusive, Shared, Invalid)
//           (4) MOESI (Modified, Owned, Exclusive, Shared, Invalid)
///////////////////////////////////////////////////////////////////////////////

#include <moesi/processor.hpp>
#include <moesi/protocol.hpp>

namespace cpptesk {
namespace examples {
namespace moesi {

Protocol* Protocol::create(CacheProtocol protocol, Core &core)
{
    switch(protocol)
    {
    case   MSI: { return new   MsiProtocol(core); }
    case  MOSI: { return new  MosiProtocol(core); }
    case  MESI: { return new  MesiProtocol(core); }
    case MOESI: { return new MoesiProtocol(core); }
    default:    { return 0;                       }
    }
}

CPPTESK_DEFINE_ACTION(Protocol::empty)
{
    CPPTESK_UNUSED_ARG(CPPTESK_GET_TRANSITION());
}

CPPTESK_DEFINE_ACTION(Protocol::error)
{
    CPPTESK_ASSERTION(false, CPPTESK_GET_TRANSITION());
}

CPPTESK_DEFINE_ACTION(Protocol::read)
{
    CPPTESK_UNUSED_ARG(CPPTESK_GET_TRANSITION());

    const size_t I = _core.get_id();
    const size_t N = _core.get_cores_number();

    for(size_t i = 0; i < N; i++)
    {
        if(i == I) { continue; }

        switch(_core.get_core(i).get_state())
        {
        case MODIFIED:  { _core.send(SN_R, i); break; }
        case OWNED:     { _core.send(SN_R, i); break; }
        case EXCLUSIVE: { _core.send(SN_R, i); break; }
        case SHARED:    { /* Do nothing */     break; }
        case INVALID:   { /* Do nothing */     break; }
        }
    }
}

CPPTESK_DEFINE_ACTION(Protocol::write)
{
    CPPTESK_UNUSED_ARG(CPPTESK_GET_TRANSITION());

    const size_t I = _core.get_id();
    const size_t N = _core.get_cores_number();

    for(size_t i = 0; i < N; i++)
    {
        if(i == I) { continue; }

        switch(_core.get_core(i).get_state())
        {
        case MODIFIED:  { _core.send(SN_I, i); break; }
        case OWNED:     { _core.send(SN_I, i); break; }
        case EXCLUSIVE: { _core.send(SN_I, i); break; }
        case SHARED:    { _core.send(SN_I, i); break; }
        case INVALID:   { /* Do nothing */     break; }
        }
    }
}

CPPTESK_DEFINE_ACTION(Protocol::evict)
{
    CPPTESK_UNUSED_ARG(CPPTESK_GET_TRANSITION());
}

CPPTESK_DEFINE_ACTION(Protocol::snoop)
{
    CPPTESK_UNUSED_ARG(CPPTESK_GET_TRANSITION());
}

}}} // cpptesk::examples::moesi

