/*
 * Copyright 2012-2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK (cpptesk::examples)
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Family of scenarios for the MOESI cache coherence protocols:
//           (1) MSI   (Modified,                   Shared, Invalid)
//           (2) MOSI  (Modified, Owned,            Shared, Invalid)
//           (3) MESI  (Modified,        Exclusive, Shared, Invalid)
//           (4) MOESI (Modified, Owned, Exclusive, Shared, Invalid)
///////////////////////////////////////////////////////////////////////////////

#include <moesi_scenario.hpp>

#include <iostream>

using namespace cpptesk::hw;

namespace cpptesk {
namespace examples {
namespace moesi {

bool MoesiScenario::init(int argc, char **argv)
{
    UNUSED_ARG(argc);
    UNUSED_ARG(argv);

    std::cout << "Simulation has started:"
              << " protocol " << CacheProtocol(MoesiConfig::get().cache_protocol)
              << " (" << MoesiConfig::get().cores_number << " cores)"
              << std::endl;

    return true;
}

void MoesiScenario::finish()
{
    std::cout << "Simulation has finished." << std::endl;
}

bool MoesiScenario::read(MoesiContext &ctx)
{
    size_t id;

    IBEGIN
    {
        for(IVAR(i) = 0; (size_t)IVAR(i) < get_number(); IVAR(i)++)
        {
            IACTION
            {
                if((id = get_core(IVAR(i))) != -1u)
                {
                    _model->read(id);
                }
                YIELD(_model->invariant());
            }
        }
    }
    IEND
}

bool MoesiScenario::write(MoesiContext &ctx)
{
    size_t id;

    IBEGIN
    {
        for(IVAR(i) = 0; (size_t)IVAR(i) < get_number(); IVAR(i)++)
        {
            IACTION
            {
                if((id = get_core(IVAR(i))) != -1u)
                {
                    _model->write(id);
                }
                YIELD(_model->invariant());
            }
        }
    }
    IEND
}

bool MoesiScenario::evict(MoesiContext &ctx)
{
    size_t id;

    IBEGIN
    {
        for(IVAR(i) = 0; (size_t)IVAR(i) < get_number(); IVAR(i)++)
        {
            IACTION
            {
                if((id = get_core(IVAR(i))) != -1u)
                {
                    _model->evict(id);
                }
                YIELD(_model->invariant());
            }
        }
    }
    IEND
}

bool MoesiScenario::reset(MoesiContext &ctx)
{
    IBEGIN
    {
        IACTION
        {
            _model->reset();
            YIELD(_model->invariant());
        }
    }
    IEND
}

////////////////////////////////////////////////////////////////////////////////
// State abstraction
////////////////////////////////////////////////////////////////////////////////

std::string MoesiScenario::get_state()
{
    std::vector<CacheState> state(_model->get_cores_number()); 

    for(size_t i = 0; i < _model->get_cores_number(); i++)
        { state[i] = _model->get_core(i).get_state(); }

    std::stringstream buffer;

    switch(MoesiConfig::get().state_abstraction)
    {
    case VECTOR:
        { buffer << VectorState  <CacheState>(state.begin(), state.end()); break; }
    case MULTISET:
        { buffer << MultisetState<CacheState>(state.begin(), state.end()); break; }
    case SET:
        { buffer << SetState     <CacheState>(state.begin(), state.end()); break; }
    }

    return buffer.str();
}

////////////////////////////////////////////////////////////////////////////////
// Stimulus abstraction
////////////////////////////////////////////////////////////////////////////////

size_t MoesiScenario::get_number() const
{
    switch(MoesiConfig::get().stimulus_abstraction)
    {
    case INDEX: { return get_indices_number(); }
    case STATE: { return get_states_number(); }
    }

    return 0;
}

size_t MoesiScenario::get_core(int stimulus) const
{
    switch(MoesiConfig::get().stimulus_abstraction)
    {
    case INDEX: { return get_core_by_index(stimulus); }
    case STATE: { return get_core_by_state(stimulus); }
    }

    return -1u;
}

size_t MoesiScenario::get_indices_number() const
{
    return MoesiConfig::get().cores_number;
}

size_t MoesiScenario::get_core_by_index(int index) const
{
    return size_t(index);
}

size_t MoesiScenario::get_states_number() const
{
    return _states_number;
}

size_t MoesiScenario::get_core_by_state(int state) const
{
    std::vector<size_t> cores;

    for(size_t i = 0; i < _model->get_cores_number(); i++)
    {
        if(_model->get_core(i).get_state() == _states[state])
            { cores.push_back(_model->get_core(i).get_id()); }
    }

    return cores.empty() ? -1u : cores[rand() % cores.size()];
}

}}} // cpptesk::examples::moesi

