/** @file fifo_media.cpp
 * This file contains the mediator class implementation.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <fifo_media.h>

#include <sync.h>

#include <interface.h>
#include <vpi_media.h>

#include <iostream>

namespace cpptesk {
namespace examples {
namespace fifo {

FIFOMediator::FIFOMediator()
{
    // Set level of debug messages
    SET_DEBUG_LEVEL(DEBUG_INFO, true);
    SET_DEBUG_STYLE(DEBUG_USER, BLACK, CYAN, CYAN, BLACK);

    // Registration of interfaces, serializers, and deserializers
    SET_INPUT_ADAPTER(iface1, FIFOMediator::serialize_iface1);
    SET_INPUT_ADAPTER(iface2, FIFOMediator::serialize_iface2);
    SET_OUTPUT_ADAPTER(iface3, FIFOMediator::deserialize_iface3);

    // Maximum reaction waiting time
    SET_REACTION_TIMEOUT(FIFOMediator::REACTION_TIMEOUT);
}

FIFOMediator::~FIFOMediator() {}

bool FIFOMediator::isIface1Ready() const
{
    return FIFO::isIface1Ready();
}
bool FIFOMediator::isIface2Ready() const
{
    return FIFO::isIface2Ready();
}

DEFINE_PROCESS(FIFOMediator::serialize_iface1)
{
    // Copy message to the local variable
    InputData msg = CAST_MESSAGE(InputData);

    // Start the serializer's working
    START_PROCESS();

    // Print this words if DEBUG_USER debug level is active
    DEBUG_PRINT(DEBUG_USER, "FIFOMediator::serialize_iface1");
    DEBUG_PRINT(DEBUG_USER, "FIFOMediator::serialize_iface1: message=" << GET_MESSAGE());

    // Set iface1 state to busy
    CAPTURE_IFACE();

    // Set values to the DUT's inputs
    inputs.CE_WR = 1;
    inputs.DI = msg.get_data();

    // Set iface1 state to free
    RELEASE_IFACE();

    // The process is stopped until the next cycle of simulation
    CYCLE();

    // Stop the serializer's working
    STOP_PROCESS();
}

DEFINE_PROCESS(FIFOMediator::serialize_iface2)
{
    // Copy message to the local variable
    InputData msg = CAST_MESSAGE(InputData);

    // Start the serializer's working
    START_PROCESS();

    // Print this words if DEBUG_USER debug level is active
    DEBUG_PRINT(DEBUG_USER, "FIFOMediator::serialize_iface2");
    DEBUG_PRINT(DEBUG_USER, "FIFOMediator::serialize_iface2: message=" << GET_MESSAGE());

    // Set iface2 state to busy
    CAPTURE_IFACE();

    // Set values to the DUT's inputs
    inputs.CE_RD = 1;

    // Set iface2 state to free
    RELEASE_IFACE();

    // The process is stopped until the next cycle of simulation
    CYCLE();

    // Stop the serializer's working
    STOP_PROCESS();
}

DEFINE_PROCESS(FIFOMediator::deserialize_iface3)
{
    // Copy message to the local variable
    OutputData &msg = CAST_MESSAGE(OutputData);

    // Start the deserializer's working
    START_PROCESS();

    // Print this words if DEBUG_USER debug level is active
    DEBUG_PRINT_FILE_LINE(DEBUG_USER, "FIFOMediator::deserialize_iface3");

    // Wait several cycles until DUT sets DO_VAL==1
    WAIT_REACTION(outputs.DO_VAL);

    // Save data sent by DUT
    msg.set_data(outputs.DO);

    // Select next registered reaction at this interface if given
    NEXT_REACTION();

    // Stop the deserializer's working
    STOP_PROCESS();
}

}}}

