/*
 * Copyright 2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * File: aspect_fsm.c
 * Description:
 *       This file contains implementation of Aspectrace aspect "fsm".
 */


#include <utils/assertion.h>
#include "aspect_fsm.h"
#include "aspectrace_internals.h"

static const char ASPECT[] = "fsm";

static const char DEFAULT_FSM_ID[] = "0";


void aspect_fsm_traceFsmStart( const char *scenarioName )
{
	if (aspectrace_allowed( ASPECT, NULL ))
		aspect_traceTagStart( ASPECT, DEFAULT_FSM_ID, NULL );

	if (aspectrace_allowed( ASPECT, "start" )) {
		const char *rootAttrs[] = { "name", DEFAULT_FSM_ID, "full-name", scenarioName, NULL };
		aspect_trace(ASPECT, NULL, create_TraceNode("start", rootAttrs, NULL, 0));
	}
}

void aspect_fsm_traceFsmEnd( )
{
	if (aspectrace_allowed( ASPECT, NULL ))
		aspect_traceTagEnd( ASPECT, DEFAULT_FSM_ID, NULL );

	if (aspectrace_allowed(ASPECT, "end")) {
		const char *rootAttrs[] = { "name", DEFAULT_FSM_ID, NULL };
		aspect_trace(ASPECT, NULL, create_TraceNode("end", rootAttrs, NULL, 0));
	}
}

void aspect_fsm_traceState( const char *stateId, const char *stateText )
{
	if (!aspectrace_allowed( ASPECT, "state" ))
		return;

	if (stateText == NULL)
		stateText = stateId;

	{
		const char *rootAttrs[] = { "id", stateId, "name", stateText, "fsm", DEFAULT_FSM_ID, NULL };
		aspect_trace(ASPECT, NULL, create_TraceNode("state", rootAttrs, NULL, 0));
	}
}

void aspect_fsm_traceTransitionStart( const char *id, const char *scenFuncName )
{
	if (aspectrace_allowed( ASPECT, NULL ))
		aspect_traceTagStart( ASPECT, id, DEFAULT_FSM_ID );

	if (aspectrace_allowed( ASPECT, "start-transition" )) {
		const char *rootAttrs[] = { "id", id, "name", scenFuncName, "fsm", DEFAULT_FSM_ID, NULL };
		aspect_trace(ASPECT, NULL, create_TraceNode("start-transition", rootAttrs, NULL, 0));
	}
}

void aspect_fsm_traceTransitionEnd( const char *id, const char *scenFuncName )
{
	if (aspectrace_allowed( ASPECT, "end-transition" )) {
		const char *rootAttrs[] = { "id", id, "name", scenFuncName, "fsm", DEFAULT_FSM_ID, NULL };
		aspect_trace(ASPECT, NULL, create_TraceNode("end-transition", rootAttrs, NULL, 0));
	}

	if (aspectrace_allowed( ASPECT, NULL ))
		aspect_traceTagEnd( ASPECT, id, DEFAULT_FSM_ID );
}

void aspect_fsm_traceIterationVariable( const char *type, const char *name, Object *value )
{
	if (!aspectrace_allowed( ASPECT, "iteration-variable" ))
		return;

	{
		const char *rootAttrs[] = { "value", NULL, "fsm", DEFAULT_FSM_ID,
			"type", type, "name", name, NULL };
		String *ss = toString_Object( value );
		rootAttrs[1] = toCharArray_String( r(ss) );

		aspect_trace(ASPECT, NULL, create_TraceNode("iteration-variable", rootAttrs, NULL, 0));

		destroy( ss );
	}
}

void aspect_fsm_traceSummary( Map *summary )
{
TraceNode **children;
int size, i;

	if (!aspectrace_allowed( ASPECT, "summary" ))
		return;

	size = size_Map( r(summary) );
	children = (TraceNode**)calloc( size+1, sizeof(TraceNode*) );
	assertion( children != NULL, FORMAT("aspect_fsm_traceSummary: No memory") );

	for (i = 0; i < size; i++) {
		String *key = key_Map( r(summary), i );
		String *val = get_Map( r(summary), r(key) );
		const char *attrs[] = {
			"name", toCharArray_String(r(key)),
			"value", toCharArray_String(r(val)),
			NULL };
		children[i] = create_TraceNode("property", attrs, NULL, 0);
		/* char arrays will persist while summary is alive */
		destroy( key );
		destroy( val );
	}
	children[size] = NULL;

	{
		const char *rootAttrs[] = { "fsm", DEFAULT_FSM_ID, NULL };
		aspect_trace(ASPECT, NULL, create_TraceNode("summary", rootAttrs, children, 0));
	}

	destroy( summary );
	free( children );
}

void aspect_fsm_traceBrowserMessage( enum MessageLevel level, const char *message )
{
	if (aspectrace_allowed( ASPECT, "browser-message" )) {
		TraceNode *children[] = { create_textTraceNode( message ), NULL };
		const char *level_str = to_str_MessageLevel( level );
		const char *rootAttrs[] = { "fsm", DEFAULT_FSM_ID, "level", level_str, NULL };

		aspect_trace(ASPECT, NULL, create_TraceNode("browser-message", rootAttrs, children, 0));
	}
}

