/*
 * Copyright 2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * File: aspectrace.c
 * Description:
 *       This file contains Aspectrace trace control routines.
 */

#include <string.h>

#include <atl/hashmap.h>
#include <atl/list.h>
#include <atl/string.h>
#include <atl/stringbuffer.h>
#include <atl/void_ast.h>
#include <utils/assertion.h>
#include <utils/network.h>
#include "aspectrace_internals.h"


static const char START_STRING[] =
	"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n"
	"<trace format=\"mixed\" version=\"3.0\" t:def-channel=\"0\">\n";
static const char FINISH_STRING[] = "</trace>\n";


/*******************************
 *         Trace nodes         *
 *******************************/

static TraceNode *alloc_tn()
{
	TraceNode *res = malloc( sizeof(TraceNode) );
	assertion( res != 0, FORMAT("create_TraceNode: No memory") );
	return res;
}

TraceNode *create_TraceNode(const char *name, const char **attrs, TraceNode **children, size_t count)
{
TraceNode *res = alloc_tn();

	res->isText = false;

#ifdef ALLOC_NEW
	res->s = strdup(name);
#else
	res->s = name;
#endif

	if (attrs != NULL) {
		int i = 0;
		res->attrs = create_Map(&type_String, &type_String);
		while (attrs[i] != NULL) {
			String *k = create_String( attrs[i++] );
			String *v = create_String( attrs[i++] );
			destroy(	// Free memory if non-NULL returned (should not be)
				put_Map( r(res->attrs), k, v )
			);
		}
	} else {
		res->attrs = NULL;
	}

	res->nextSibling = NULL;

	if (children == NULL || count == 0) {
		res->firstChild = NULL;
	} else {
		TraceNode *current = *children;
		res->firstChild = current;
		while (--count) {
			current->nextSibling = *(++children);
			current = current->nextSibling;
		}
        current->nextSibling = NULL;
	}

	return res;
}

size_t list_len(const void** list)
{
    if (NULL == list) return 0;
    size_t result = 0;
    for(;*list != NULL;list++)
    {
        result ++;
    }
    return result;
}

// TraceNode *create_TraceNode_a(const char *name, const char **attrs, TraceNode **children)
// {
//     return create_TraceNode(name, attrs, children, list_len((const void**)children));
// }

TraceNode *create_textTraceNode(const char *text)
{
TraceNode *res = alloc_tn();

#ifdef ALLOC_NEW
	res->s = (text == NULL)? "" : strdup(text);
#else
	res->s = (text == NULL)? "" : text;
#endif

	res->isText = true;
	res->attrs = NULL;
	res->nextSibling = NULL;
	res->firstChild = NULL;

	return res;
}

TraceNode *create_stringTraceNode(String *text)
{
/* TODO: Special handling for case ALLOC_NEW */
TraceNode *res;

	res = create_textTraceNode( toCharArray_String(r(text)) );
	destroy(text);
	return res;
}

void destroy_TraceNode(TraceNode *node)
{
#ifdef ALLOC_NEW
	free( node->s );
#endif

	destroy( node->attrs );

	while (node->firstChild != NULL)	{
		TraceNode *next = node->firstChild->nextSibling;
		destroy_TraceNode( node->firstChild );
		node->firstChild = next;
	}
	// nextSibling is cleared by parent

	free(node);
}


/*******************************
 *        Trace channels       *
 *******************************/

typedef struct TraceChannel
{
	void *p;
	void (*traceString)(const char *s, int len, void *p);
	void (*close)(void *p);
	struct TraceChannel *next;
} TraceChannel;

static TraceChannel *channels;

static TraceChannel *alloc_tc()
{
TraceChannel *res = malloc( sizeof(TraceChannel) );
	assertion( res != 0, FORMAT("create_TraceChannel: No memory") );
	return res;
}

static void add_channel(TraceChannel *channel)
{
	channel->next = NULL;
	if (channels == NULL) {
		channels = channel;
	} else {
		TraceChannel *t = channels;
		while (t->next != NULL)
			t = t->next;
		t->next = channel;
	}

	channel->traceString(START_STRING, -1, channel->p);
}

static void multiplex_chararray(const char *s, int len)
{
TraceChannel *t = channels;

	while (t != NULL) {
		t->traceString(s, len, t->p);
		t = t->next;
	}
}

static void multiplex(const char *s)
{
	multiplex_chararray(s, -1);
}

void aspectrace_finishTracing()
{
	multiplex(FINISH_STRING);

	while (channels != NULL) {
		TraceChannel *next = channels->next;
		channels->close( channels->p );
		free( channels );
		channels = next;
	}
}


/*********** Trace to file ***********/

static void traceToFile(const char *s, int len, void *p)
{
	if (len < 0)
	{
		fputs(s, (FILE*)p);
	}
	else if (len == 0)
	{
		return;
	}
	else
	{
		fwrite(s, len, 1, (FILE*)p);
	}
}

static void closeFile(void *p)
{
	fclose((FILE*)p);
}

void aspectrace_addXmlTrace(const char *fileName)
{
FILE *file = fopen(fileName, "wt");
TraceChannel *tc;

	file = fopen(fileName, "wt");
	if (file == NULL) {
		fprintf(stderr, "aspectrace_addXmlTrace: Failed to create trace %s", fileName);
		return;
	}

	tc = alloc_tc();
	tc->p = file;
	tc->traceString = traceToFile;
	tc->close = closeFile;

	add_channel(tc);
}


/*********** Trace to sockets ***********/

static void sendToSocket(const char *s, int len, void *p)
{
int sock = *(int*)p;
	if (len <0) len = strlen(s);
	send_to_socket(sock, s, len);
}

static void closeSocket(void *p)
{
int sock = *(int*)p;
	close_socket(sock);
	free(p);
}

void aspectrace_addTcpSocket(const char *serverHost, unsigned short serverPort)
{
int sock;
TraceChannel *tc;

	sock = connect_to_server(serverHost, serverPort);
	if (sock < 0)	return;

	tc = alloc_tc();
	tc->traceString = sendToSocket;
	tc->close = closeSocket;

	tc->p = malloc(sizeof(int));
	assertion( tc->p != 0, FORMAT("aspectrace_addTcpSocket: No memory") );
	*(int*)tc->p = sock;

	add_channel(tc);
}


/*******************************
 * Trace formatting (XML only) *
 *******************************/

/*********** Messages ***********/

static char *indent_buffer = NULL;
static int buffer_len = 0;
static int spaces = 0;

static void increase_indent() {
	spaces += 2;
	if (spaces >= buffer_len) {
		buffer_len += 16;
		indent_buffer = realloc(indent_buffer, buffer_len);	// indent_buffer may be NULL before call
		assertion( indent_buffer != 0, FORMAT("aspect_trace: No memory") );
	}
	strcpy(indent_buffer + spaces - 2, "  ");
}

static void decrease_indent() {
	spaces -= 2;
	indent_buffer[ spaces ] = 0;
}

static void multiplex_sb( StringBuffer *sb )
{
	String *s = toString(sb);
	multiplex( toCharArray_String(r(s)) );
	destroy(s);
}

static void appendString(StringBuffer* sb, const char * str)
{
	int start = 0;
	while (str[start])
	{
		size_t pos = strcspn(str+start, "<>&\r\n\"\r\n");
		appendCharArray_StringBuffer(r(sb), str+start, pos);
		const char * rep = NULL;
		switch (str[start+pos])
		{
			case '"': rep = "&quot;"; break;
			case '<': rep = "&lt;"; break;
			case '>': rep = "&gt;"; break;
			case '&': rep = "&amp;"; break;
			case '\r': rep = "&#x0D;"; break;
			case '\n': rep = "&#x0A;"; break;
			case '\0': destroy(sb); return;
		}
		append_StringBuffer(r(sb), rep);
		start += pos + 1;
	}
	destroy(sb);
}

static void printNode(TraceNode *n, const char *aspect /*, const char *channel*/)
{
int i;
TraceNode *curNode;
StringBuffer *sb;
	if (channels == NULL)
		return;

	if (n->isText) {
		// TODO If the string contains ']]>' sequence it must be split into sequences ]] and >
		multiplex("<![CDATA[");
		const char * start = n->s;
		const char * pos = start;
		while(pos != NULL)
		{
			pos = strstr(start, "]]>");
			if (pos == NULL)
			{
				multiplex(start);
			}
			else
			{
				const char * end = pos+2;
				multiplex_chararray(start, end-start);
				multiplex("]]><![CDATA[");
				start = end;
			}
		}
//		multiplex(n->s);
		multiplex("]]>");
		multiplex("\n");
		return;
	}

	if (indent_buffer != NULL)
		multiplex(indent_buffer);

	sb = create_StringBuffer();
	append_StringBuffer( r(sb), "<" );
	if (aspect != NULL) {
		appendString( r(sb), aspect );
		append_StringBuffer( r(sb), ":" );
	}
	appendString( r(sb), n->s );
/*
	if (channel != NULL) {
		appendFormat_StringBuffer( r(sb), " t:ch=\"%s\"", channel );
	}
*/

	if (n->attrs != NULL) {
		for (i = 0; i < size_Map(r(n->attrs)); i++ ) {
			String *k = key_Map( r(n->attrs), i );
			String *v = get_Map( r(n->attrs), r(k) );
			append_StringBuffer( r(sb), " ");
			append_StringBuffer( r(sb), toCharArray_String(r(k)));
			append_StringBuffer( r(sb), "=\"");

			appendString(r(sb), toCharArray_String(r(v)) );
			append_StringBuffer( r(sb), "\"");
			destroy(k);
			destroy(v);
		}
	}

	curNode = n->firstChild;
	if (curNode == NULL) {
		append_StringBuffer( r(sb), "/>\n" );
		multiplex_sb( sb );
		return;
	}

	append_StringBuffer( r(sb), ">\n");
	multiplex_sb( r(sb) );
	clear_StringBuffer( r(sb) );

	increase_indent();
	while (curNode != NULL) {
		printNode( curNode, NULL /*, NULL*/ );
		curNode = curNode->nextSibling;
	}
	decrease_indent();

	multiplex(indent_buffer);	// non-NULL here

	append_StringBuffer( r(sb), "</");
	if (aspect != NULL) {
		appendString( r(sb), aspect );
		append_StringBuffer( r(sb), ":" );
	}
	appendString( r(sb), n->s );
	append_StringBuffer( r(sb), ">\n");
	multiplex_sb( sb );
	return;
}

void aspect_trace(const char *aspect, const char *channel, TraceNode *node)
{
	assertion( ! node->isText, "aspect_trace: text node on top level" );

	if (!aspectrace_allowed(aspect, node->s))
		return;

	printNode( node, aspect /*, channel*/ );
	destroy_TraceNode( node );
}

/*********** Tags ***********/

static void traceTagStartEnd(const char *action, const char *aspect, const char *tag, const char *parent)
{
StringBuffer *sb;

	if (!aspectrace_allowed(aspect, NULL))
		return;

	if (indent_buffer != NULL)
		multiplex(indent_buffer);

	sb = create_StringBuffer();
	appendFormat_StringBuffer( r(sb), "<tag-%s aspect=\"%s\" name=\"%s\"", action, aspect, tag );
	if (parent != NULL)
		appendFormat_StringBuffer( r(sb), " parent=\"%s\"", parent );
	append_StringBuffer( r(sb), "/>\n" );

	multiplex_sb( sb );
}

void aspect_traceTagStart(const char *aspect, const char *tag, const char *parent)
{
	traceTagStartEnd("start", aspect, tag, parent);
}

void aspect_traceTagEnd  (const char *aspect, const char *tag, const char *parent)
{
	traceTagStartEnd("end", aspect, tag, parent);
}


/*******************************
 *       Trace filtering       *
 *******************************/

static List *filters = NULL;

/*static int hash_VoidAst( VoidAst *va )
{
	void * ptr = value_VoidAst( va );
	int i = (int)ptr;

	return i ^ (i << 4) ^ (i >> 8);
}*/

void aspectrace_addFilter(AspectraceFilter filter)
{
VoidAst *va = create_VoidAst( filter );

	if (filters == NULL)
		filters = create_List( &type_VoidAst );

	if (contains_List( r(filters), r(va) )) {
		destroy( va );
		return;
	}

	append_List( r(filters), va );
}

void aspectrace_removeFilter(AspectraceFilter filter)
{
int idx;

	if (filters == NULL)	return;

	idx = indexOf_List( r(filters), create_VoidAst(filter) );
	if (idx >= 0)
		remove_List( r(filter), idx );
}

void aspectrace_removeAllFilters()
{
	if (filters == NULL)	return;

	clear_List( r(filters) );
}


/************ Aspect filtering ************/

static HashMap *allowed_aspects = NULL;

static bool aspect_filter(const char *aspect, const char *message)
{
	return allowed_aspects==NULL
		|| containsKey_HashMap( r(allowed_aspects), create_String(aspect) );
}


void aspectrace_setAllowedAspects(const char *aspects, ...)
{
va_list args;
const char *ptr;

	if (allowed_aspects == NULL)
		 allowed_aspects = create_HashMap( &type_String, &type_String, 32, (HashFunc)hash_String );
	else clear_HashMap( r(allowed_aspects) );

	if (aspects != NULL)
		put_HashMap( r(allowed_aspects), create_String(aspects), NULL );

	va_start( args, aspects );
	for (;;) {
		ptr = va_arg( args, const char * );
		if (ptr == NULL)	break;
		put_HashMap( r(allowed_aspects), create_String(ptr), NULL );
	}
	va_end( args );

	aspectrace_addFilter(aspect_filter);
}

void aspectrace_setAllowedAspects_a(int count, const char **aspects)
{
int i;

	if (allowed_aspects == NULL)
		 allowed_aspects = create_HashMap( &type_String, &type_String, 32, (HashFunc)hash_String );
	else clear_HashMap( r(allowed_aspects) );

	for (i = 0; i < count; i++) {
		put_HashMap( r(allowed_aspects), create_String(aspects[i]), NULL );
	}

	aspectrace_addFilter(aspect_filter);
}

bool aspectrace_allowed(const char *aspect, const char *message /*NULL for tag*/)
{
int i, len;

	if (filters == NULL)	return true;

	len = size_List( r(filters) );
	for (i = 0; i < len; i++) {
		VoidAst *va = get_List( r(filters), i );
		AspectraceFilter f = (AspectraceFilter) value_VoidAst( va );
		if (! f(aspect,message) )
			return false;
	}
	return true;
}


/*******************************
 *         Trace levels        *
 *******************************/

const char *to_str_MessageLevel( enum MessageLevel level )
{
	switch (level) {
	case ML_FATAL:
		return "FATAL";
	case ML_ERROR:
		return "ERROR";
	case ML_WARNING:
		return "WARNING";
	case ML_INFO:
		return "INFO";
	case ML_DEBUG:
		return "DEBUG";
	default:
		assertion( false, FORMAT( "Invalid message level" ) );
	}
	return NULL;
}
