#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <limits.h>
#include <getopt.h>

#ifndef LINE_MAX
#define LINE_MAX 254
#endif

#if defined(WIN32)
  #include <windows.h>
#elif defined(unix)
#include <sys/time.h>
  #include <pthread.h>
#else
  #error Unknown OS!
#endif

#include <utils/network.h>
#include <atl/string.h>
#include <atl/stringbuffer.h>
#include <ts/system.h>

#include "synchronizer/synchronizer.h"
#include "netndfsm.h"
#include "command.h"
#include "agent.h"

/******************************************************************************/
/**                                NodeStatus                                **/
/******************************************************************************/
typedef enum NodeStatus {
      NODE_INITIAL
    , NODE_RUN
    , READING_CFG
    , TEST_READY
    , TEST_STARTED
    , TEST_STOPPED
} NodeStatus;

String *toString_NodeStatus( NodeStatus status )
{
    switch(status) {
        case NODE_INITIAL: return create_String( "NODE_INITIAL" );
        case NODE_RUN:     return create_String( "NODE_RUN" );
        case READING_CFG:  return create_String( "READING_CFG" );
        case TEST_READY:   return create_String( "TEST_READY" );
        case TEST_STARTED: return create_String( "TEST_STARTED" );
        case TEST_STOPPED: return create_String( "TEST_STOPPED" );
        default:           return format_String( "unknown(%d)", status );
    }
}

/******************************************************************************/
/**                                   Agent                                  **/
/******************************************************************************/
typedef struct Agent {
    TE_Graph *graph;
    TestEngine *te;
    TestScenario *scen;
    Synchronizer *sync;
    int argc;
    char **argv;

    unsigned short port;
    int socket;
    int connection;

    int cpu_id;
    String *id;

    NodeStatus status;

#if defined(WIN32)
    HANDLE thread_TestEngine;
#elif defined(unix)
    pthread_t thread_TestEngine;
#endif

} Agent;


char *get_timemark()
{
    struct timeval tp;
    struct tm *tm;
    static char timemark[40];

    gettimeofday(&tp, NULL);
    tm = localtime(&(tp.tv_sec));

    sprintf(timemark, "[%d:%02d:%02d.%06ld]", tm->tm_hour, tm->tm_min, tm->tm_sec, (long)tp.tv_usec);

    return timemark;
}

/******************************************************************************/
/**                              Debug printf                                **/
/******************************************************************************/
#if __STDC_VERSION__ >= 199901L
#ifdef DEBUG_AGENT
void debug_printf(char* format, ...)
{
    va_list arg_list;
    int res = 0;

    printf("%s ", get_timemark());

    va_start(arg_list, format);
    vprintf(format, arg_list);
    va_end(arg_list);

    fflush(stdout);
}
#else
#define debug_printf(...) /* nothing */
#endif
#else
void debug_printf(char* format, ...)
{
#ifdef DEBUG_AGENT
    va_list arg_list;
    int res = 0;

    printf("%s ", get_timemark());

    va_start(arg_list, format);
    vprintf(format, arg_list);
    va_end(arg_list);

    fflush(stdout);
#endif
    return;
}
#endif


/******************************************************************************/
/**                           Test Engine thread                             **/
/******************************************************************************/
static bool verdict;
static void *thread_TestEngine( void *par )
{
    Agent *agent = (Agent *) par;

    assertion(agent != NULL, "thread_TestEngine: agent == NULL");
    debug_printf("AGENT: TestEngine thread started.\n");

    verdict = main_TestEngine( agent->te, agent->graph, agent->scen, agent->argc, agent->argv );

    debug_printf("AGENT: TestEngine thread finished.\n");
    return &verdict;
}


/*** start_TestEngineThread ***/
#if defined(WIN32)

static DWORD WINAPI ThreadTestEngine(LPVOID lpParam)
{
    thread_TestEngine( (Agent*)lpParam );
    ExitThread(0);
    return 0;
}
int start_TestEngineThread( Agent *agent )
{
    DWORD dwThreadId;

    if (agent->thread_TestEngine != 0) {
        fprintf( stderr, "TestEngine thread already started!\n" );
        return -1;
    }

    agent->thread_TestEngine = CreateThread(NULL, 0, ThreadTestEngine, thread_TestEngine, 0, &dwThreadId);
    if (agent->thread_TestEngine == 0) {
        printf("Creating of TestEngine thread failed: %d\n", GetLastError());fflush(stdout);
        return -1;
    }

    return 0;
}

#elif defined(unix)

int start_TestEngineThread( Agent *agent )
{
    assertion(agent != NULL, "start_TestEngineThread: agent == NULL");
    if( agent->thread_TestEngine != 0 ) {
        fprintf( stderr, "TestEngine thread already started!\n" );
        return -1;
    }

    if( pthread_create( &agent->thread_TestEngine, NULL, thread_TestEngine, agent) != 0 ) {
        perror( "pthread_create" );
        return -1;
    }

    return 0;
}
#endif


int stop_TestEngineThread( Agent *agent )
{
int res;

    debug_printf("AGENT: stop_TestEngineThread...\n");
    assertion(agent != NULL, "stop_TestEngineThread: agent == NULL");
    if( agent->thread_TestEngine == 0 )
        return 0;

    debug_printf("AGENT: stop_TestEngineThread: before finish_TestEngine ...\n");
    finish_TestEngine( agent->te, agent->graph );
    debug_printf("AGENT: stop_TestEngineThread: after finish_TestEngine.\n");

#if defined(WIN32)
    for (;;) {
        DWORD exitCode;
        res = ! GetExitCodeThread( agent->thread_TestEngine, &exitCode );
        if (exitCode == STILL_ACTIVE)
            sleep_msec( 10 );
        else break;
    }
    res = ! CloseHandle( agent->thread_TestEngine ) || res;    // 0 iff CloseHandle()==true && GetExitCodeThread==true
#elif defined(unix)
    res = pthread_join( agent->thread_TestEngine, NULL );
    if (res != 0)
        perror( "pthread_join" );
#endif

    agent->thread_TestEngine = 0;

    debug_printf("AGENT: stop_TestEngineThread done.\n");
    return res;
}

/******************************************************************************/
/**                              create_Agent                                **/
/******************************************************************************/
void create_Agent( Agent *agent, unsigned short port,
                   TE_Graph *graph, TestEngine *te,
                   TestScenario *scen,  Synchronizer *sync,
                   int argc, char **argv )
{
    bool initialized = false;

    assertion(agent != NULL, "create_Agent: agent == NULL");
    agent->port = port;
    agent->graph = graph;
    agent->te = te;
    agent->scen = scen;
    agent->sync = sync;
    agent->argc = argc;
    agent->argv = argv;

    agent->cpu_id = -1;
    agent->id = NULL;

    agent->status = NODE_RUN;

    agent->thread_TestEngine = 0;

    debug_printf("AGENT: Agent has been created.\n");
}

/******************************************************************************/
/**                              start_Agent                                 **/
/******************************************************************************/
static int set_configuration( Agent *agent );
static int start( Agent *agent );
static int stop( Agent *agent );
static int send_info( Agent *agent );

int start_Agent( Agent *agent )
{
    debug_printf("AGENT: Starting Agent ...\n");
    assertion(agent != NULL, "start_Agent: agent == NULL");

    agent->socket = create_server( agent->port );
    if( agent->socket == -1 ) return -1;
    agent->connection = -1;

    while( agent->status != TEST_STOPPED )
    {
        String *strNodeStatus = toString_NodeStatus( agent->status );
        int rc;
        char buffer[AGENT_MSG_SIZE];


        debug_printf("AGENT: State %s. Waiting for coordinator on port %d ...\n", toCharArray_String( r(strNodeStatus) ), agent->port);
        destroy(strNodeStatus);
        if(agent->connection == -1) {
            agent->connection = accept_connection( agent->socket, true );
        }

        memset( buffer, 0, AGENT_MSG_SIZE );
        rc = read_from_socket( agent->connection, buffer, AGENT_MSG_SIZE );
        if( rc == 0 ) {
            // mcc has closed connections
             close_socket( agent->connection );
             agent->connection = -1;
        } else {
            debug_printf("AGENT: <- from coordinator: %d <%s>\n", rc, buffer);
            if( strcmp( buffer, MCC_CMD_CFG ) == 0 )
            {
                if( agent->status == NODE_RUN) {
                    rc = set_configuration( agent );
                    if( !rc ) agent->status = TEST_READY;
                } else {
                    /* Configuration is already set */
                    rc = send_to_socket( agent->connection, AGENT_ANSWER_FAIL AGENT_ERROR_CFG, AGENT_MSG_SIZE );
                    debug_printf("AGENT: ->: %d <%s>\n", rc, AGENT_ANSWER_FAIL AGENT_ERROR_CFG);
                }
            }
            else if( strcmp( buffer, MCC_CMD_START ) == 0 )
            {
                if( agent->status == TEST_READY ) {
                    rc = start( agent );
                    if( !rc ) agent->status = TEST_STARTED;
                } else {
                    /* Already started */
                    rc = send_to_socket( agent->connection, AGENT_ANSWER_FAIL AGENT_ERROR_STARTED, AGENT_MSG_SIZE );
                    debug_printf("AGENT: ->: %d <%s>\n", rc, AGENT_ANSWER_FAIL AGENT_ERROR_STARTED );
                }
            }
            else if( strcmp( buffer, MCC_CMD_STOP ) == 0 )
            {
                if( agent->status == TEST_STARTED ) {
                    rc = stop( agent );
                    if( !rc ) agent->status = TEST_STOPPED;
                } else {
                    /* Not started */
                    rc = send_to_socket( agent->connection, AGENT_ANSWER_FAIL AGENT_ERROR_NOTSTARTED, AGENT_MSG_SIZE );
                    debug_printf("AGENT: ->: %d <%s>\n", rc, AGENT_ANSWER_FAIL AGENT_ERROR_NOTSTARTED );
                }
            }
            else if( strcmp( buffer, MCC_CMD_INFO ) == 0 )
            {
                rc = send_info( agent );
            }
            else
            {
                rc = send_to_socket( agent->connection, AGENT_ANSWER_UNKNOWN, AGENT_MSG_SIZE );
                debug_printf("AGENT: ->: %d <%s>\n", rc, AGENT_ANSWER_UNKNOWN );
            }
        }
    }

    debug_printf("AGENT: start(Agent):done\n");
    return 0;
}

/******************************************************************************/
/**                          set_configuration                               **/
/******************************************************************************/
static int set_configuration( Agent *agent ) {
    int number;
    unsigned short port;
    int numberOfInConnections;
    int rc;
    char buffer[AGENT_MSG_SIZE];
    String* attr, *number_str;
    int saveLog, loadLog;
    char logFileName[AGENT_MSG_SIZE];
    char id[AGENT_MSG_SIZE];

    assertion(agent != NULL, "set_configuration: agent == NULL");
    rc = read_from_socket( agent->connection, buffer, AGENT_MSG_SIZE );
    sscanf( buffer, MCC_CMD_CFG_INFO, &number, &port, &numberOfInConnections, id );

    debug_printf("AGENT: Configuration: number:%d, port:%d, number of in connections:%d, id:%s\n", number, port, numberOfInConnections, id);

    attr = create_String(TE_ATTR_CPU_ID);
    number_str = format_String("%d", number);
    setAttrValue_TestEngine(agent->te, agent->graph, attr, number_str);
    destroy(attr);
    destroy(number_str);

    agent->cpu_id = number;
    agent->id = create_String( id );

    rc = sync_SetInputConf( agent->sync, port, numberOfInConnections );
    if( rc ) {
        /* can't configure Synchronizer (in connection) */
        send_to_socket( agent->connection, AGENT_ANSWER_FAIL AGENT_ERROR_SYNC_INCONF, AGENT_MSG_SIZE );
        return -1;
    }

    rc = read_from_socket( agent->connection, buffer, AGENT_MSG_SIZE );
    while( strcmp(buffer, MCC_CMD_CFG_END) != 0 ) {
        char address[AGENT_MSG_SIZE];
        unsigned short port;

        sscanf( buffer, MCC_CMD_CFG_OUT, &port, address );
        debug_printf("AGENT: Configuration: out connection: address:<%s>, port:%d\n", address, port);
        rc = sync_AddOutConnection( agent->sync, address, port );

        rc = read_from_socket( agent->connection, buffer, AGENT_MSG_SIZE );
    }

    sync_TolerateNeighbourDeaths( agent->sync, true );

    rc = read_from_socket( agent->connection, buffer, AGENT_MSG_SIZE );
	debug_printf("AGENT: Configuration: <%s>", buffer);
    sscanf( buffer, MCC_CMD_CFG_LOG, &saveLog, &loadLog, logFileName );
    debug_printf("AGENT: Configuration: logging: save:%d, load:%d, fileName:<%s>\n",
                 saveLog, loadLog, logFileName
                 );

    if( loadLog ) {
        rc = sync_LoadLog( agent->sync, logFileName, saveLog, false );
        debug_printf("AGENT: Configuration: loading log from <%s>, rc = %d\n", logFileName, rc);
    } else if( saveLog ) {
        rc = sync_SaveLog( agent->sync, logFileName );
        debug_printf("AGENT: Configuration: logging to <%s>, rc = %d\n", logFileName, rc);
    }


    send_to_socket( agent->connection, AGENT_ANSWER_OK MCC_CMD_CFG, AGENT_MSG_SIZE );

    return 0;
}

/******************************************************************************/
/**                                  start                                   **/
/******************************************************************************/
static int start( Agent *agent )
{
    int rc;

    assertion(agent != NULL, "start: agent == NULL");
    send_to_socket( agent->connection, AGENT_ANSWER_OK MCC_CMD_START, AGENT_MSG_SIZE );

    rc = sync_ConfigurationFinished( agent->sync );
    if( rc ) {
        /* can't configure Synchronizer (out connections) */
        debug_printf("AGENT: !!! start(Agent): Can't configure Synchronizer (out connections): rc = %d !!!\n", rc);
        debug_printf("AGENT: start(Agent): <%s> ...\n", AGENT_ANSWER_FAIL AGENT_ERROR_SYNC_OUTCONF);
//        send_to_socket( agent->connection, AGENT_ANSWER_FAIL AGENT_ERROR_SYNC_OUTCONF, AGENT_MSG_SIZE );
        return -1;
    }

    debug_printf("AGENT: start(Agent):Starting Synchronizer thread...\n");
    rc = sync_StartThread( agent->sync );
    if( rc ) {
        /* can't start Synchronizer */
//        send_to_socket( agent->connection, AGENT_ANSWER_FAIL AGENT_ERROR_SYNC_START, AGENT_MSG_SIZE );
        return -1;
    }

    debug_printf("AGENT: start(Agent):Starting TestEngineThread...\n");
    rc = start_TestEngineThread( agent );

    if( rc ) {
        debug_printf("AGENT: start(Agent):can't start testEngineThread\n");
        /* can't start TestEngine */
//        send_to_socket( agent->connection, AGENT_ANSWER_FAIL AGENT_ERROR_TE_START, AGENT_MSG_SIZE );
        return -1;
    }

//    send_to_socket( agent->connection, AGENT_ANSWER_OK MCC_CMD_START, AGENT_MSG_SIZE );
    debug_printf("AGENT: start(Agent):done\n");
    return 0;
}


/******************************************************************************/
/**                                   stop                                   **/
/******************************************************************************/
static int stop( Agent *agent )
{
    int rc;

    debug_printf("AGENT: stop(Agent) ...\n");
    assertion(agent != NULL, "stop: agent == NULL");

    send_to_socket( agent->connection, AGENT_ANSWER_OK MCC_CMD_STOP, AGENT_MSG_SIZE );

    debug_printf("AGENT: stop(Agent): before sync_FinishThread...\n");
    rc = sync_FinishThread( agent->sync );
    debug_printf("AGENT: stop(Agent): after sync_FinishThread.\n");

    if( rc ) {
        /* can't stop Synchronizer */
//        send_to_socket( agent->connection, AGENT_ANSWER_FAIL AGENT_ERROR_SYNC_STOP, AGENT_MSG_SIZE );
        debug_printf("AGENT: stop(Agent): ERROR: can't stop Synchronizer\n");
        //return -1;
    }

    debug_printf("AGENT: stop(Agent): before stop_TestEngineThread...\n");
    rc = stop_TestEngineThread( agent );
    debug_printf("AGENT: stop(Agent): after stop_TestEngineThread.\n");

    if( rc ) {
        /* can't stop TestEngine */
//        send_to_socket( agent->connection, AGENT_ANSWER_FAIL AGENT_ERROR_TE_STOP, AGENT_MSG_SIZE );
        debug_printf("AGENT: stop(Agent): ERROR: can't stop TestEngine\n");
        //return -1;
    }

    debug_printf("AGENT: stop(Agent) done.\n");
    return 0;
}

/******************************************************************************/
/**                                send_info                                 **/
/******************************************************************************/
static String *to_JSON_String_NodeStatus( NodeStatus status )
{
    String *res;

    switch(status) {
        case NODE_INITIAL: res = create_String("initial"); break;
        case NODE_RUN:     res = create_String("node run"); break;
        case READING_CFG:  res = create_String("reading configuration"); break;
        case TEST_READY:   res = create_String("test ready"); break;
        case TEST_STARTED: res = create_String("test started"); break;
        case TEST_STOPPED: res = create_String("test stopped"); break;
        default:           res = create_String("unknown"); break;
    }

    return res;
}


typedef struct cpu_info {
    unsigned long long cpu;
    unsigned long long nice;
    unsigned long long system;
    unsigned long long idle;
    unsigned long long busy;
    unsigned long long work;
    double usage;
} CPU_info;

static String *getCpuInfo()
{
    StringBuffer *buf  = create_StringBuffer();

    static CPU_info info[10] = {}; // info[0] - total, info[1] ... info[9] - cpu(i-1)
    
    FILE *f;
    char line[LINE_MAX];
    int i, line_number = 0, multi_cpu = 0;
    
    if( f = fopen( "/proc/stat", "r" ) ) {
        while( fgets ( line, LINE_MAX, f ) != NULL ) {
            if( line_number == 0 ) {
                if( sscanf( line, "cpu %llu %llu %llu %llu", &info[0].cpu, &info[0].nice, &info[0].system, &info[0].idle ) < 4 ) break;
            } else {
                int cpu_number;
                if( sscanf( line, "cpu%d %llu %llu %llu %llu", &cpu_number,
                            &info[line_number].cpu, &info[line_number].nice, &info[line_number].system, &info[line_number].idle ) < 5 ) break;
            }
            line_number++;
        }
        fclose( f );
        
        // check if there is more than one cpu
        if( line_number > 2 ) multi_cpu = 1;

        for( i = 0; i < line_number; i++ ) {
            unsigned long long busy = info[i].cpu + info[i].nice + info[i].system;
            unsigned long long work = busy + info[i].idle;

            info[i].usage = 100.0 * ( busy - info[i].busy ) / ( double ) ( work - info[i].work );

            if ( i > 1 ) append_StringBuffer( r(buf), ", " );
            appendFormat_StringBuffer( r(buf), "%.2f%%", info[i].usage );
            if ( i == 0 && multi_cpu ) append_StringBuffer( r(buf), " (" );

            info[i].busy = busy;
            info[i].work = work;

            if ( !multi_cpu ) break;
        }
        if ( multi_cpu ) append_StringBuffer( r(buf), ")" );
    }

    return toString_Object(buf);;
}

static String *getMemInfo()
{
    StringBuffer *buf  = create_StringBuffer();

    FILE *f;
    char line[LINE_MAX];
    unsigned long long memTotal, memFree;
    double memUsage;

    if( f = fopen( "/proc/meminfo", "r" ) ) {
        // MemTotal
        if( fgets( line, LINE_MAX, f ) != NULL ) {
            if( sscanf( line, "MemTotal: %llu", &memTotal ) < 1 ) memTotal = 0;
        }
        
        // MemFree
        if( fgets( line, LINE_MAX, f ) != NULL ) {
            if( sscanf ( line, "MemFree: %llu", &memFree ) < 1 ) memFree = 0;
        }
        
        fclose ( f );

        memUsage = 100.0 * (memTotal - memFree) / ( double ) memTotal;
        appendFormat_StringBuffer( r(buf), "%.2f%%/%.2fGb", memUsage, ( double ) memTotal / ( double ) 1024 / ( double ) 1024 );       
    }

    return toString_Object(buf);;
}

static String *getAgentInfo( Agent *agent )
{
    StringBuffer *buf  = create_StringBuffer();
    String *str_status;
    String *cpu, *mem;

    str_status = to_JSON_String_NodeStatus( agent->status );
    append_StringBuffer( r(buf), "{" );

    append_StringBuffer( r(buf), "\"status\":\"" );
    appendString_StringBuffer( r(buf), r(str_status) );
    append_StringBuffer( r(buf), "\"" );
    append_StringBuffer( r(buf), "," );

    append_StringBuffer( r(buf), "\"id\":\"" );
    agent->id == NULL? append_StringBuffer( r(buf), "" ) : appendString_StringBuffer( r(buf), r(agent->id) );
    append_StringBuffer( r(buf), "\"" );
    append_StringBuffer( r(buf), "," );

    append_StringBuffer( r(buf), "\"timemark\":\"" );
    appendFormat_StringBuffer( r(buf), "%s", get_timemark());
    append_StringBuffer( r(buf), "\"" );
    append_StringBuffer( r(buf), "," );

    cpu = getCpuInfo();
    append_StringBuffer( r(buf), "\"cpu\":\"" );
    appendString_StringBuffer( r(buf), r(cpu) );
    append_StringBuffer( r(buf), "\"" );
    append_StringBuffer( r(buf), "," );

    mem = getMemInfo();
    append_StringBuffer( r(buf), "\"memory\":\"" );
    appendString_StringBuffer( r(buf), r(mem) );
    append_StringBuffer( r(buf), "\"" );

    append_StringBuffer( r(buf), "}" );

    destroy(str_status);
    destroy(cpu);
    destroy(mem);

    return toString_Object(buf);
}

static String *Map_to_JSON_String( Map *map )
{
    int i;
    StringBuffer* buf  = create_StringBuffer();

    append_StringBuffer( r(buf), "{" );
    if( map != NULL ) {
        for( i = 0; i < size_Map( r(map) ); i++ ) {
            String *key = (String *)key_Map( r(map), i );
            String *val = (String *)get_Map( r(map), r(key) );

            if( i > 0 ) append_StringBuffer( r(buf), "," );
            append_StringBuffer( r(buf), "\"" );
            appendString_StringBuffer( r(buf), key);
            append_StringBuffer( r(buf), "\":\"" );
            appendString_StringBuffer( r(buf), val);
            append_StringBuffer( r(buf), "\"" );

        }
    }
    append_StringBuffer( r(buf), "}" );

    return toString_Object(buf);
}

static int send_info( Agent *agent )
{
    int rc, len;
    Map *te_attrs, *sync_attrs;
    String *te_info, *sync_info, *agent_info, *info;
    StringBuffer *buf  = create_StringBuffer();
    String *tmp;

    assertion(agent != NULL, "send_info: agent == NULL");

    debug_printf("AGENT: getAllAttrValues_TestEngine\n");
    te_attrs = getAllAttrValues_TestEngine( agent->te, agent->graph );
    te_info = Map_to_JSON_String( te_attrs );

    debug_printf("AGENT: getAllAttrValues_Synchronizer\n");
    sync_attrs = getAllAttrValues_Synchronizer( agent->sync );
    sync_info = Map_to_JSON_String( sync_attrs );

    debug_printf("AGENT: getAgentInfo\n");
    agent_info = getAgentInfo( agent );

    append_StringBuffer( r(buf), "{" );
    append_StringBuffer( r(buf), "\"TestEngine\":" );
    appendString_StringBuffer( r(buf), r(te_info) );
    append_StringBuffer( r(buf), ", \"Synchronizer\":" );
    appendString_StringBuffer( r(buf), r(sync_info) );
    append_StringBuffer( r(buf), ", \"Agent\":" );
    appendString_StringBuffer( r(buf), r(agent_info) );
    append_StringBuffer( r(buf), "}" );

    info = toString_Object( r(buf) );

    /* send length of string */
    len = length_String( r(info) );
    tmp = valueOfInt_String( len );

    send_to_socket( agent->connection,
                    toCharArray_String( r(tmp) ),
                    AGENT_MSG_SIZE );
    /* send string itself*/
    send_to_socket( agent->connection, toCharArray_String( r(info) ), len + 1 );

    destroy( te_attrs );
    destroy( sync_attrs );
    destroy( te_info );
    destroy( sync_info );
    destroy( agent_info );
    destroy( info );
    destroy( buf );
    destroy( tmp );

    return 0;
}

/******************************************************************************/
/**                                 netmain                                  **/
/******************************************************************************/
int netmain(TestEngine *engine, TestScenario *scen, int argc, char **argv)
{
    TE_Graph graph;
    Synchronizer *sync;
    Agent ag;
    unsigned short port = DEFAULT_AGENT_PORT;
    int rc;
    sigset_t signal_mask;

 /* command line options */
    const char* short_options = "p:";
    const struct option long_options[] = {
        {"port", required_argument, NULL, 'p'},
        {NULL,   0,                 NULL, 0}
    };
    int rez, option_index;
    extern int opterr;

    opterr = 0; // prevent error messages
    while( (rez = getopt_long( argc, argv, short_options,
                               long_options, &option_index )) != -1 ){

        switch( rez ){
            case 'p': {
                port = atoi( optarg );
                debug_printf("AGENT: netmain: port = %d(%s)\n",port, optarg);
                break;
            }
        }
    }

#ifdef CREATE_STRING_SEMAPHORE
    // Create a semaphore for string data
    CREATE_STRING_SEMAPHORE();
	debug_printf("AGENT: create string semaphore\n");
#endif

#if defined(unix)
    // Ignore SIGPIPE
    sigemptyset( &signal_mask );
    sigaddset(&signal_mask, SIGPIPE);
    rc = pthread_sigmask(SIG_BLOCK, &signal_mask, NULL);
#endif

    debug_printf("AGENT: netmain: init_Graph ...\n");
    init_Graph(&graph, TE_UPDATE_ON, &TE_DEFAULT_MUTEX);

    debug_printf("AGENT: netmain: init_TestEngine ...\n");
    init_TestEngine(engine, &graph, scen, argc, argv);

    debug_printf("AGENT: netmain: create_Synchronizer ...\n");
    sync = create_Synchronizer( &graph, engine );

    debug_printf("AGENT: netmain: create_Agent ...\n");
    create_Agent(&ag, port, &graph, engine, scen, sync, argc, argv);

    debug_printf("AGENT: netmain: start_Agent ...\n");
    rc = start_Agent(&ag);

    destroy_Synchronizer(sync);
    destroy_Graph(&graph);

    debug_printf("AGENT: netmain done.\n");
    return 0;
}


