/*
 * Copyright 2010-2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <netfsm.h>
#include <sgl/graph.h>
#include <sgl/utils.h>
#include <utils/assertion.h>

//------------------------------------------------------------------------------
// FSM Macro
//------------------------------------------------------------------------------

#define FSM_FIND_PATH_TRIES    10
#define FSM_RANDOM_EDGES_TRIES 100

//------------------------------------------------------------------------------
// FSM Error Codes
//------------------------------------------------------------------------------

typedef enum FSM_Errno
{
    /// No errors.
    FSM_ERRNO_NO_ERRORS = 0x0,
    /// Unconnected graph.
    FSM_ERRNO_UNCONNECT = 0x1
} FSM_Errno;

//------------------------------------------------------------------------------
// FSM Data
//------------------------------------------------------------------------------

typedef struct FSM_Data
{
    /// The reference to the external graph information.
    TE_GraphInfo *info;

    /// The command-line parameters.
    int    argc;
    char **argv;

    /// Try nondeterministic edges if there are no others.
    int nondeterm;
    /// Randomize choice of an untraversed edge.
    int randomize;

    /// The previous number of the locally traversed edges (to track endless loops).
    int edges_traversed_locally;
    /// The number of resultless find-path tries.
    int no_new_edge_count;
    /// The count of randomly generatod stimuli to break out of the loop.
    int random_edge_count;
    /// The index of the node in which the loop has been detected.
    int start_node_index;
    /// The next edge index.
    int next_edge_index;
    /// Choose arcs randomly.
    int random_edge_choice;
} FSM_Data;

static FSM_Data* alloc_Data(void)
{
    FSM_Data *data = (FSM_Data*)malloc(sizeof(FSM_Data));
    assertion(data != NULL, "alloc_Data: data == NULL");
    
    return data;
}

static void dealloc_Data(FSM_Data *data)
{
    int i;

    if(data == NULL) { return; }

    if(data->argv != NULL)
    {
        for(i = 0; i < data->argc; i++)
            { free(data->argv[i]); }
            
        free(data->argv);
    }
    
    free(data);
}

static void init_Data(FSM_Data *data, TE_GraphInfo *info)
{
    data->info = info;
    
    data->argc = 0;
    data->argv = NULL;

    data->nondeterm  = 0;
    data->randomize  = 0;

    data->edges_traversed_locally =  0;
    data->no_new_edge_count       =  0;
    data->random_edge_count       = -1;
    data->start_node_index        = -1;
    data->next_edge_index         =  0;
    data->random_edge_choice      =  0;
}

//------------------------------------------------------------------------------
// FSM Interface
//------------------------------------------------------------------------------

static void processParams_FSM(TE_Graph *graph, int argc, char **argv)
{
    int i;

    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    FSM_Data  *data2 = (FSM_Data*) data1->data;

    assertion(argc >= 0,                 "processParams_FSM: argc < 0");
    assertion(argv != NULL || argc == 0, "processParams_FSM: argv == NULL && argc > 0");
    
    if(argc > 0)
        { data2->argv = (char**)malloc(argc * sizeof(char*)); }
    
    for(i = 0; i < argc; i++)
    {
        if(!strcmp("--cpu", argv[i]))
        {
            assertion(++i < argc, "processParams_FSM: --cpu should be followed by an integer CPU identifier");
            data2->info->cpu_id = atoi(argv[i]);

            sgl_setId_Graph(data1, data2->info->cpu_id);
        }
        else if(!strcmp("--nondeterministic", argv[i]))
        {
            data2->nondeterm = 1;
        }
        else if(!strcmp("--randomize", argv[i]))
        {
            data2->randomize = 1;
        }

        data2->argv[data2->argc++] = strdup(argv[i]);
    }
}

static void init_FSM(TE_Graph *graph, int argc, char **argv)
{
    SGL_Graph *data1 = sgl_alloc_Graph();
    FSM_Data  *data2 = alloc_Data();

    sgl_init_Graph(data1, NULL, NULL, (SGL_DataDestructor)dealloc_Data);
    init_Data(data2, &graph->info);

    data1->data = data2;
    graph->data = data1;

    processParams_FSM(graph, argc, argv);
}

static void config_FSM(TE_Graph *graph, const char *attr_name, const char *attr_value)
{
    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    FSM_Data  *data2 = (FSM_Data*) data1->data;

    if(!strcmp(TE_ATTR_CPU_ID, attr_name))
    {
        data2->info->cpu_id = atoi(attr_value);
        sgl_setId_Graph(data1, data2->info->cpu_id);
    }
    else
    {
        assertion(false, "config_FSM: unknown attribute");
    }
}

static TE_UpdateStatus start_FSM(TE_Graph *graph, TE_NodeId *node_id)
{
    SGL_UpdateStatus status;

    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    FSM_Data  *data2 = (FSM_Data*)data1->data;

    sgl_addNode_Graph(data1, node_id, &status);

    data2->info->nodes_traversed++;
    data2->info->nodes_traversed_locally++;

    return status == SGL_US_OLD_ALL ? TE_OLD_ALL : TE_NEW_NODE;
}

static TE_UpdateStatus update_FSM(TE_Graph *graph, TE_UpdateType type,
    TE_ArcId *arc_id, TE_NodeId *new_end_node_id)
{
    SGL_UpdateStatus status;

    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    FSM_Data  *data2 = (FSM_Data*) data1->data;

    SGL_NodeId *end_node_id = new_end_node_id != NULL ? new_end_node_id : arc_id->end_node_id;
    SGL_Edge *edge = sgl_addEdge_Graph(data1, arc_id->start_node_id, arc_id->stimulus_id, end_node_id, &status);

    if((status & SGL_US_NEW_EDGE) != 0)
    {
        SGL_Node *start_node = sgl_getStartNode_Edge(edge);
#if 0
        SGL_EdgeChecker edge_checker = data2->nondeterm ?
            sgl_emptyEdgeChecker : sgl_deterministicEdgeChecker;

        sgl_updateWeight_Node(start_node, sgl_nonFinishedNodeChecker, edge_checker);
#else
        if(!sgl_isFinished_Node(start_node))
            { sgl_setWeight_Node(start_node, SGL_MINIMAL_WEIGHT); }
#endif
    }

    return updateInfo_Edge(data2->info, edge, type, status);
}

static bool isTrap_FSM(TE_Graph *graph, SGL_Node *node)
{
    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    FSM_Data  *data2 = (FSM_Data*) data1->data;

    // If the number of locally traversed edges does not change for a long,
    // the situation is considered to be a loop or some other trap.
    if(data2->edges_traversed_locally == data2->info->arcs_traversed_locally)
    {
        data2->no_new_edge_count++;
    }
    else
    {
        data2->no_new_edge_count =  0;
        data2->random_edge_count = -1;
    }

    data2->edges_traversed_locally = data2->info->arcs_traversed_locally;

    if(data2->no_new_edge_count >= FSM_FIND_PATH_TRIES)
    {
        if(data2->random_edge_count < 0)
        {
            data2->random_edge_count  = FSM_RANDOM_EDGES_TRIES;
            data2->random_edge_choice = sgl_next_Random(data1) & 1;
            data2->next_edge_index    = sgl_next_Random(data1);
            data2->start_node_index   = node->id;
            
            return true;
        }
        
        if(data2->random_edge_count == 0 || node->id != data2->start_node_index)
        {
            data2->no_new_edge_count = 0;
            data2->random_edge_count = 0;
            
            return false;
        }
	
        data2->random_edge_count--;

        return true;
    }

    return false;
}

static TE_ArcList* getPath_FSM(TE_Graph *graph, TE_NodeId *node_id)
{
    int i;

    TE_ArcList *path;

    SGL_Node *node = NULL;
    SGL_Edge *edge = NULL;

    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    FSM_Data  *data2 = (FSM_Data*) data1->data;

    SGL_EdgeChecker edge_checker = data2->nondeterm ?
        sgl_emptyEdgeChecker : sgl_deterministicEdgeChecker;

    // Traversal has been completed.
    if(data2->info->arcs_traversed >= data2->info->arcs_known)
    {
        graph->error = FSM_ERRNO_NO_ERRORS;
        return NULL;
    }

    // Get the current node.
    node = sgl_findNode_Graph(data1, node_id);
    assertion(node != NULL, "getPath_FSM: node == NULL");

    // Check for a trap.
    if(!isTrap_FSM(graph, node))
    {
        // The graph is unconnected.
        if(!sgl_findPath_NonFinishedNode_Graph(data1, node, edge_checker, data2->randomize))
        {
            graph->error = FSM_ERRNO_UNCONNECT;
            return NULL;
        }

        // Convert the path into the external format.
        path = create_List(&type_TE_ArcId);
        for(i = data1->path.size - 1; i >= 0; i--)
        {
            edge = *((SGL_Edge**)get_ArrayList(&data1->path, i));
            node = sgl_getEndNode_Edge(edge, 0);

            append_List(r(path), createArcId_Edge(edge));

            if(!sgl_isDeterministic_Edge(edge))
                { node = NULL; break; }
        }
    }
    else
    {
        // Try to break out of the trap.
        path = create_List(&type_TE_ArcId);
    }

    if(node != NULL)
    {
        if(sgl_getDegree_Node(node) > 0)
        {
            if(sgl_isFinished_Node(node))
            {
                if(data2->random_edge_choice)
                    { edge = sgl_getEdge_Random(data1, node); }
                else
                    { edge = sgl_getEdge_Node(node, data2->next_edge_index++ % sgl_getDegree_Node(node)); }
            }
            else
                { edge = sgl_getNewEdge_Random(data1, node); }

            append_List(r(path), createArcId_Edge(edge));
        }
    }

    return path;
}

static TE_ArcList* reachNode_FSM(TE_Graph *graph, TE_NodeId *node_id, TE_ArcId *arc_id)
{
    int i;
    TE_ArcList *path;

    SGL_Node *start_node;
    SGL_Node *target_node;
    
    SGL_Edge *edge = NULL;

    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    FSM_Data  *data2 = (FSM_Data*) data1->data;

    SGL_EdgeChecker edge_checker = data2->nondeterm ?
        sgl_emptyEdgeChecker : sgl_deterministicEdgeChecker;

    // Get the initial node.
    start_node = sgl_getNode_Graph(data1, 0);
    assertion(start_node != NULL, "reachNode_FSM: start_node == NULL");

    // Get the target node.
    target_node = sgl_findNode_Graph(data1, node_id);
    assertion(target_node != NULL, "reachNode_FSM: target_node == NULL");

    // Try to find the path from the initial node to the target one.
    if(!sgl_findPath_TargetNode_Graph(data1, start_node, target_node, edge_checker, data2->randomize))
        { return NULL; }

    // Convert the path into the external format.
    path = create_List(&type_TE_ArcId);
    for(i = data1->path.size - 1; i >= 0; i--)
    {
        edge = *((SGL_Edge**)get_ArrayList(&data1->path, i));
        start_node = sgl_getEndNode_Edge(edge, 0);

        append_List(r(path), createArcId_Edge(edge));

        if(!sgl_isDeterministic_Edge(edge))
            { start_node = NULL; break; }
    }

    if(start_node != NULL && arc_id != NULL)
        { append_List(r(path), r(arc_id)); }

    return path;
}

static String* getError_FSM(TE_Graph *graph)
{
    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    FSM_Data  *data2 = (FSM_Data*) data1->data;

    switch(graph->error)
    {
    case FSM_ERRNO_NO_ERRORS:
        {
            return create_String("None");
        }
    case FSM_ERRNO_UNCONNECT:
        {
            if(data1 != NULL)
            {
                SGL_Node *node = sgl_findNonFinishedNode_Graph(data1);

                if(node != NULL)
                {
                    bool nd_graph = data2->info->arcs_nondeterministic > 0;
                    bool nd_allow = data2->nondeterm;

                    const char *error_reason =  nd_graph ?
                    (nd_allow ? "too much nondeterministic behavior" :
                                "nondeterministic behavior with no '--nondeterministic' option set") :
                                "incomplete set of actions or too strong preconditions";

                    return format_String("Unconnected graph: nonfinished state '%s' could not be reached (%s)",
                        NODE_ID_STRING(sgl_getNodeId_Node(node)), error_reason);
                }
            }

            return create_String("Unconnected graph");
        }
    default:
        {
            return create_String("Unknown error");
        }
    }
}

static bool toFile_FSM(TE_Graph *graph, const char *filename)
{
    SGL_Graph *data1 = (SGL_Graph*)graph->data;
    FSM_Data  *data2 = (FSM_Data*) data1->data;
    
    return sgl_toFile_Graph(data1, filename, data2->argc, data2->argv);
}

static void finish_FSM(TE_Graph *graph)
{
    SGL_Graph *data1 = (SGL_Graph*)graph->data;

    if(data1 != NULL)
    {
        FSM_Data *data2 = (FSM_Data*)data1->data;

        sgl_destroy_Graph(data1);
        sgl_dealloc_Graph(data1);

        graph->data = NULL;
    }
}

//------------------------------------------------------------------------------
// FSM Engine
//------------------------------------------------------------------------------

TestEngine TE_FSM =
{
    /* .name        = */ "netfsm",
    /* .init        = */ init_FSM,
    /* .config      = */ config_FSM,
    /* .start       = */ start_FSM,
    /* .update      = */ update_FSM,
    /* .getPath     = */ getPath_FSM,
    /* .reachNode   = */ reachNode_FSM,
    /* .getError    = */ getError_FSM,
    /* .finish      = */ finish_FSM,
    /* .toFile      = */ toFile_FSM,
    /* .fromFile    = */ NULL,
    /* .generalizes = */ NULL,
    /* .sysdata     = */ { NULL }
};
